#include "NestStep.h"
#include "Combine.h"
#include "Combine_send_recv.h"
#include <stdlib.h>
#include <mpi.h>
#include <stdio.h>
#include <string.h>

CombineItem * srbuffer;

/*
 * new_CombineInfo
 * initializes a combineinfo item and returns a pointer
 */
CombineInfo * new_CombineInfo(ShVar * var, 
							  int combine_strategy, 
							  void (*combineFn)(void *, void *, void *))
{
	CombineInfo * tempinfo;
	
	tempinfo = (CombineInfo *)malloc(sizeof(CombineInfo));
	tempinfo->var = var;
	tempinfo->combine_strategy = combine_strategy;
	tempinfo->combineFn = combineFn;
	
	return tempinfo;
}

/*
 * free_CombineInfo
 * frees the memory used by a CombineInfo item
 */
void free_CombineInfo(CombineInfo * info)
{
	free(info);
}

int clear_CombineList(CombineList * _clist)
{
	int i, j;
	CombineList * temp = _clist;
	for(i = 0; i < _clist->no_rep_arrays; ++i)
	{
		if(_clist->rep_info[i].prefix == PREFIX)
		{
			for(j = 0; j < thisgroup->tree->no_children; ++j)
				free(_clist->rep_info[i].child_arrays[j]);
			free(_clist->rep_info[i].child_arrays);
			free(_clist->rep_info[i].old_base);
		}
		_clist->rep_info[i].rep_array->url = 0;
		_clist->rep_info[i].rep_array->urh = _clist->rep_info[i].rep_array->size - 1;
	}
	// clear all
	temp->no_combine_items = 0;
	temp->no_prefix_vars = 0;
	temp->no_rep_arrays = 0;
	temp->largest_rep_array = 0;
	temp->no_block_arrays = 0;
	temp->no_cyclic_arrays = 0;
	temp->request_id_counter = 0;
	for(i = 0; i < NestStep_get_size(); ++i)
		temp->N_DAmsgs[i] = 0;

	return 0;
}

/*
 * NestStep_combine
 * sends data specified in the combinelist to the process' leader and then
 * receives updated data
 */
int NestStep_combine(CombineList * _clist, ArrayHolder * _holder)
{
	int i;
#ifdef DA_DEBUG
	int k;
#endif
	Name name;
	RepArray * N_DAmsgs;
	CombineList * combineList;
	ArrayHolder * holder;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;
	if(_holder == NULL)
		holder = arrayHolder;
	else
		holder = _holder;
	
	// create replicated array for N_DAmsgs and attach it
	name.procedure = -1;
	name.relative = -1;
	N_DAmsgs = new_RepArray(name, IVAR, NestStep_get_size(), combineList->N_DAmsgs);
	memset(N_DAmsgs->hidden, 0, NestStep_get_size() * sizeof(int));
	NestStep_RepArray_attach(N_DAmsgs,
							 ADD,
							 NULL,
							 NO_PREFIX,
							 NULL,
							 NOT_UPDATED, 
							 combineList);

	// receive data from children
	combine_receive_ch(combineList);

	// send data to parent
	combine_send_p(combineList);

	// receive data from parent
	combine_receive_p(combineList);

	// send data to children
	combine_send_ch(combineList);

#ifdef DA_DEBUG
	printf("%d: N_DAmsgs:", NestStep_get_rank());
	for(k = 0; k < NestStep_get_size(); ++k)
		printf(" %d", ((int *)(N_DAmsgs->base))[k]);
	printf("\n");
#endif

	process_DA_requests(combineList, holder);
	process_DA_incoming(combineList, holder);

	for(i = 0; i < combineList->no_rep_arrays; ++i)
	{
		int offset = combineList->rep_info[i].rep_array->url * get_element_size(combineList->rep_info[i].rep_array->type);
		int elements = (combineList->rep_info[i].rep_array->urh - combineList->rep_info[i].rep_array->url + 1);
		// copy to hidden
		memcpy((char *)combineList->rep_info[i].rep_array->hidden + offset,
			(char *)combineList->rep_info[i].rep_array->base + offset,
			elements * get_element_size(combineList->rep_info[i].rep_array->type));
	}

	clear_CombineList(combineList);
	// clear buffers here
	free_RepArray(N_DAmsgs);
	srbuffer = NULL;
	return 0;
}

/*
 * combine_send_p
 * sends messages with shared variables, replicated arrays and
 * distributed array requests
 */
int combine_send_p(CombineList * _clist)
{
	int i;
#ifdef DEBUG
	int k;
#endif

	
	if(thisgroup->tree->parent != -1)
	{
		// pack items
		for(i = 0; i < _clist->no_combine_items; ++i)
		{
			pack_send_CombineItem(_clist->combine_items[i]);
		}
		// pack arrays
		for(i = 0; i < _clist->no_rep_arrays; ++i)
		{
#ifdef DEBUG
			printf("%d: sending replicated array %d:%d to parent -", 
					NestStep_get_rank(), 
					_clist->rep_info[i].rep_array->name.procedure,
					_clist->rep_info[i].rep_array->name.relative);
			for(k = 0; k < _clist->rep_info[i].rep_array->size; ++k)
			{
				if(_clist->rep_info[i].rep_array->type == IVAR)
					printf(" %d", ((int *)(_clist->rep_info[i].rep_array->base))[k]);
				else
					printf(" %.2f", ((double *)(_clist->rep_info[i].rep_array->base))[k]);
			}
			printf("\n");
#endif
			if(_clist->rep_info[i].combine_strategy == LEADER)
				;
			else
			{	
				pack_send_RepArray(_clist->rep_info[i].rep_array);
			}
		}
		// send data
		pack_send_buffer();
		create_type_send();
        send_data(thisgroup->tree->parent, COMBINE, thisgroup->t_group.comm);
		clear_send();
	}

	return 0;
}

/*
 * combine_receive_p
 * receives data from parent
 */
int combine_receive_p(CombineList * _clist)
{
	int i;
	CombineInfo * info;
	PrefixVar * prefix;
	MPI_Status status;
	RepArray temp_array;
	
	if(thisgroup->tree->parent != -1)
	{
		for(i = 0; i < _clist->no_combine_items + _clist->no_prefix_vars; i++)
			req_recv_CombineItem();
		
		// receive replicated arrays from parent
		for(i = 0; i < _clist->no_rep_arrays; ++i)
		{
			// receive array
			req_recv_RepArray(_clist->rep_info[i].rep_array);
			
			// take care of prefix
			if(_clist->rep_info[i].prefix == PREFIX)
			{
				temp_array.name.procedure = -1;
				temp_array.name.relative = i;
				temp_array.type = _clist->rep_info[i].rep_array->type;
				temp_array.size = _clist->rep_info[i].rep_array->size;
				temp_array.base = _clist->rep_info[i].prefix_base;
				req_recv_RepArray(&temp_array);
			}	
		}
		create_type_recv();
		recv_data(thisgroup->tree->parent, COMMIT, thisgroup->t_group.comm, &status);
		srbuffer = (CombineItem *)recv_buffer;
		unpack_recv_buffer();
		clear_recv();

		for(i = 0; i < _clist->no_rep_arrays; ++i)
		{
			// take care of prefix
			if(_clist->rep_info[i].prefix == PREFIX)
			{
				merge_rep_arrays(_clist->rep_info[i].prefix_base, _clist->rep_info[i].old_base, 
									_clist->rep_info[i].type, _clist->rep_info[i].rep_array->size, 
									_clist->rep_info[i].combine_strategy, _clist->rep_info[i].combineFn, 0);
			}
		}
		// process combine items
		for(i = 0; i < _clist->no_combine_items + _clist->no_prefix_vars; ++i)
		{
			info = lookup_info(srbuffer[i], _clist);
			if(srbuffer[i].prefix_tag == UPDATE)
			{
				prefix = lookup_prefix(srbuffer[i], _clist);
				prefix->from_parent = srbuffer[i].val;
				// update own prefix variable
				// merge_values(&(prefix->my_value), srbuffer[i].val, prefix->my_value,
				//			 prefix->shared->type, info->combine_strategy, info->combineFn);
				// change the value in the buffer
				// srbuffer[i].val = prefix->my_value;
				// update my real prefix
				to_real_value(prefix->prefix, prefix->shared->type, prefix->from_parent);
			}
			else if(srbuffer[i].prefix_tag == NOT_UPDATED)
			{
				;
			}
			else
			{
				// update real shared variable
				to_real_value(info->var->base, info->var->type, srbuffer[i].val);
				to_real_value(info->var->hidden, info->var->type, srbuffer[i].val);
			}
		}
	}

	return 0;
}

int combine_send_ch(CombineList * _clist)
{
	int i, j, k;
	CombineInfo * info;
	PrefixVar * prefix;
	int element_size;
	RepArray temp_array;

	if(thisgroup->tree->parent == -1)
	{
		srbuffer = (CombineItem *)recv_buffer;
		// commit buffer to real variables and pack them
		for(i = 0; i < _clist->no_combine_items; ++i)
		{
			info = lookup_info(_clist->combine_items[i], _clist);
			srbuffer[i] = _clist->combine_items[i];
			to_real_value(info->var->base, info->var->type, _clist->combine_items[i].val);
			to_real_value(info->var->hidden, info->var->type, _clist->combine_items[i].val);
		}

		// add prefix items to the buffer, make sure they are my_value, then commit
		for(i = 0; i < _clist->no_prefix_vars; ++i)
		{
			memset(_clist->prefix_vars[i].prefix, 0, get_element_size(_clist->prefix_vars[i].shared->type));
			memset(&(srbuffer[i + _clist->no_combine_items].val), 0, sizeof(double));
			srbuffer[i + _clist->no_combine_items].name = _clist->prefix_vars[i].shared->name;
			srbuffer[i + _clist->no_combine_items].contributor = NestStep_get_rank();
			srbuffer[i + _clist->no_combine_items].prefix_tag = UPDATE;
		}
	}
	for(i = 0; i < thisgroup->tree->no_children; ++i)
	{
		// update prefixes with previous children
		for(j = 0; j < _clist->no_combine_items + _clist->no_prefix_vars; ++j)
		{
			if(srbuffer[j].prefix_tag == UPDATE)
			{
				info = lookup_info(srbuffer[j], _clist);
				prefix = lookup_prefix(srbuffer[j], _clist);
				if(i > 0)
				{
					merge_values(&(srbuffer[j].val), srbuffer[j].val, prefix->child_values[i-1],
						prefix->shared->type, info->combine_strategy, info->combineFn);
				}
				else
				{
					merge_values(&(srbuffer[j].val), srbuffer[j].val, prefix->my_value,
						prefix->shared->type, info->combine_strategy, info->combineFn);
				}
			}
			pack_send_CombineItem(srbuffer[j]);
		}
		// take care of replicated arrays
		for(int j = 0; j < _clist->no_rep_arrays; ++j)
		{
			element_size = get_element_size(_clist->rep_info[j].type);
			pack_send_RepArray(_clist->rep_info[j].rep_array);
			if(_clist->rep_info[j].prefix == PREFIX)
			{
				// merge with values from previous children then send
				if(i == 0)
				{
					if(thisgroup->tree->parent == -1)
					{
						memcpy(_clist->rep_info[j].prefix_base, _clist->rep_info[j].old_base,
								element_size * _clist->rep_info[j].rep_array->size);
					}
					else
					{
						memcpy(_clist->rep_info[j].old_base, _clist->rep_info[j].prefix_base, 
							element_size * _clist->rep_info[j].rep_array->size);
					}
				}
				else
				{
					merge_rep_arrays(_clist->rep_info[j].old_base,_clist->rep_info[j].child_arrays[i - 1],
										_clist->rep_info[j].type, _clist->rep_info[j].rep_array->size, 
										_clist->rep_info[j].combine_strategy, _clist->rep_info[j].combineFn, 0);
				}

#ifdef DEBUG
				printf("%d: sending prefix_array to child %d:", NestStep_get_rank(), i);
				for(k = 0; k < _clist->rep_info[j].rep_array->size; ++k)
				{
					if(_clist->rep_info[j].type == DVAR)
						printf(" %.2f", ((double *)(_clist->rep_info[j].old_base))[k]);
				}
				printf("\n");
#endif
				temp_array.name.procedure = -1;
				temp_array.name.relative = j;
				temp_array.type = _clist->rep_info[j].rep_array->type;
				temp_array.size = _clist->rep_info[j].rep_array->size;
				temp_array.base = _clist->rep_info[j].old_base;
				pack_send_RepArray(&temp_array);
			}
		}
		pack_send_buffer();
		create_type_send();
		send_data(thisgroup->tree->children[i], COMMIT, thisgroup->t_group.comm);
		clear_send();
	}

	return 0;
}

/*
 * combine_receive_ch
 * receives messages from the children
 */
int combine_receive_ch(CombineList * _clist)
{
	int i, j, n;
	CombineInfo * info;
	PrefixVar * prefix;
	MPI_Status status;
	int element_size;

#ifdef DEBUG
	int k;
#endif

	
	if(thisgroup->tree->no_children > 0)
	{
		// initialize for retreive
		for(i = 0; i < _clist->no_combine_items; i++)
			req_recv_CombineItem();
		for(i = 0; i < _clist->no_rep_arrays; i++)
		{
			if(_clist->rep_info[i].combine_strategy != LEADER)
				req_recv_RepArray(_clist->rep_info[i].rep_array);
		}
		create_type_recv();

		// receive from children sequentially
		for(n = 0; n < thisgroup->tree->no_children; n++)
		{
			MPI_Probe(MPI_ANY_SOURCE, COMBINE, thisgroup->t_group.comm, &status);
			for(i = 0; i < thisgroup->tree->no_children; i++)
			{
				if(thisgroup->tree->children[i] == status.MPI_SOURCE)
					break;
			}
			// retreive data
			recv_data(thisgroup->tree->children[i], COMBINE,
				thisgroup->t_group.comm, &status);
			
			srbuffer = (CombineItem *)recv_buffer;
			for(j = 0; j < _clist->no_combine_items; ++j)
			{
				// take care of prefix
				if(srbuffer[j].prefix_tag == PREFIX)
				{
					prefix = lookup_prefix(srbuffer[j], _clist);
					prefix->child_values[i] = srbuffer[j].val;
				}

				// take care of standard
				info = lookup_info(srbuffer[j], _clist);
				/*if(srbuffer[j].prefix_tag == UPDATED)
				{
					_clist->combine_items[j].val = srbuffer[j].val;
					_clist->combine_items[j].prefix_tag = UPDATED;
					info->updated = UPDATED;
				}
				else if(srbuffer[j].prefix_tag == NOT_UPDATED)
					;
					*/
				if(info->combine_strategy == UPDATED)
				{
					ShVar * var = info->var;
					if(var->type == IVAR) {
						int * real = (int *)(var->base);
						int * hidden = (int *)(var->hidden);
						if(*hidden != srbuffer[j].val.i)
							*real = srbuffer[j].val.i;
						else
							srbuffer[j].val.i = *real;
					}
					else if(var->type == FVAR) {
						float * real = (float *)(var->base);
						float * hidden = (float *)(var->hidden);
						if(*hidden != srbuffer[j].val.f)
							*real = srbuffer[j].val.f;
						else
							srbuffer[j].val.f = *real;
					}
					else if(var->type == DVAR) {
						double * real = (double *)(var->base);
						double * hidden = (double *)(var->hidden);
						if(*hidden != srbuffer[j].val.d)
							*real = srbuffer[j].val.d;
						else
							srbuffer[j].val.d = *real;
					}
					_clist->combine_items[j].val = srbuffer[j].val;
					_clist->combine_items[j].prefix_tag = UPDATED;
					info->updated = UPDATED;
				}
				else
				{
					merge(info, j, _clist);
					to_real_value(info->var->base, 
						info->var->type, 
						_clist->combine_items[i].val);
				}
			} // end for
			for(j = 0; j < _clist->no_rep_arrays; ++j)
			{
				element_size = get_element_size(_clist->rep_info[j].type);
#ifdef DEBUG
				printf("%d: before processing replicated array %d:%d from child %d -", 
						NestStep_get_rank(), 
						_clist->rep_info[j].rep_array->name.procedure,
						_clist->rep_info[j].rep_array->name.relative,
						i);
				for(k = 0; k < _clist->rep_info[j].rep_array->size; ++k)
				{
					if(_clist->rep_info[j].rep_array->type == IVAR)
						printf(" %d", ((int *)(_clist->rep_info[j].rep_array->base))[k]);
					else
						printf(" %.2f", ((double *)(_clist->rep_info[j].rep_array->base))[k]);
				}
				printf("\n");
#endif
				if(_clist->rep_info[j].combine_strategy == LEADER)
				{
					;
				}
				else
				{		
					// copy array if prefix
					if(_clist->rep_info[j].prefix == PREFIX)
					{
						memcpy(_clist->rep_info[j].child_arrays[i], 
							get_rep_start(_clist->rep_info[j].rep_array), 
							element_size * _clist->rep_info[j].rep_array->size);
					}

#ifdef DEBUG
					printf("%d: Received replicated array %d:%d from child %d -", 
							NestStep_get_rank(), 
							_clist->rep_info[j].rep_array->name.procedure, 
							_clist->rep_info[j].rep_array->name.relative,
							i);
					for(k = 0; k < _clist->rep_info[j].rep_array->size; ++k)
					{
						if(_clist->rep_info[j].rep_array->type == IVAR)
							printf(" %d", ((int *)get_rep_start(_clist->rep_info[j].rep_array))[k]);
						else
							printf(" %.2f", ((double *)get_rep_start(_clist->rep_info[j].rep_array))[k]);
					}
					printf("\n");
#endif
					merge_rep_arrays((char *)_clist->rep_info[j].rep_array->base + _clist->rep_info[j].rep_array->url * element_size, 
									 get_rep_start(_clist->rep_info[j].rep_array),
									 _clist->rep_info[j].type, 
									 _clist->rep_info[j].rep_array->urh - _clist->rep_info[j].rep_array->url + 1,
									 _clist->rep_info[j].combine_strategy, _clist->rep_info[j].combineFn,
									 (char *)_clist->rep_info[j].rep_array->hidden + _clist->rep_info[j].rep_array->url * element_size);
				} // end else
#ifdef DEBUG
				printf("%d: after processing replicated array from child %d -", NestStep_get_rank(), i);
				for(k = 0; k < _clist->rep_info[j].rep_array->size; ++k)
				{
					if(_clist->rep_info[j].rep_array->type == IVAR)
						printf(" %d", ((int *)(_clist->rep_info[j].rep_array->base))[k]);
					else
						printf(" %.2f", ((double *)(_clist->rep_info[j].rep_array->base))[k]);
				}
				printf("\n");
#endif
			}// end for
		} // end for
		clear_recv();
	}// end if

	return 0;
}

/*
 * lookup_info
 * returns a pointer to the info item associated with the combine item
 * returns NULL otherwise
 */
CombineInfo * lookup_info(CombineItem item, CombineList * _clist)
{
	CombineInfo * info;
	CombineInfo tempinfo;
	ShVar * tempvar;

	tempvar = new_ShVar(item.name, IVAR, NULL); // type does not matter here
	tempinfo.var = tempvar;

	info = (CombineInfo *)bsearch((const void *)&tempinfo, 
								  (const void *)_clist->combine_info,
								  _clist->no_combine_items,
								  sizeof(CombineInfo),
								  compare_info);

	free_ShVar(tempvar);

	return info;
}

/*
 * lookup_prefix
 * returns a pointer to the prefix item in the prefix array
 * returns NULL if it doesn't exist
 */
PrefixVar * lookup_prefix(CombineItem item, CombineList * _clist)
{
	PrefixVar * prefix;
	PrefixVar tempprefix;
	ShVar * tempvar;

#ifdef SEARCH_DEBUG
	int i, k;
	printf("%d:%d looking up prefix: %d:%d:",
		NestStep_get_group_id(), NestStep_get_rank(),
		item.name.procedure, item.name.relative);
	for(k = 0; item.name.path[k] != -1; k++)
		printf("%d", item.name.path[k]);
	printf("\n");
	printf("%d:%d in list:",
		NestStep_get_group_id(), NestStep_get_rank());
	for(i = 0; i < _clist->no_prefix_vars; i++)
	{
		printf(" %d:%d:", 
			_clist->prefix_vars[i].shared->name.procedure,
			_clist->prefix_vars[i].shared->name.relative);
		for(k = 0; _clist->prefix_vars[i].shared->name.path[k] != -1; k++)
			printf("%d", _clist->prefix_vars[i].shared->name.path[k]);
	}
	printf("\n");
#endif

	tempvar = new_ShVar(item.name, IVAR, NULL); // type does not matter here
	tempprefix.shared = tempvar;

	prefix = (PrefixVar *)bsearch((const void *)&tempprefix,
								  (const void *)_clist->prefix_vars,
								  _clist->no_prefix_vars,
								  sizeof(PrefixVar),
								  compare_prefix);
	free_ShVar(tempvar);

	return prefix;
}

/*
 * compare_info
 * compares info objects
 */
int compare_info(const void * _info1, const void * _info2)
{
	const CombineInfo * info1 = (const CombineInfo *)_info1;
	const CombineInfo * info2 = (const CombineInfo *)_info2;

	return compare_name(info1->var->name, info2->var->name);
}

/*
 * compare_prefix
 */
int compare_prefix(const void * _prefix1, const void * _prefix2)
{
	const PrefixVar * prefix1 = (const PrefixVar *)_prefix1;
	const PrefixVar * prefix2 = (const PrefixVar *)_prefix2;

	return compare_name(prefix1->shared->name, prefix2->shared->name);
}


int merge(CombineInfo * _info, int _i, CombineList * _clist)
{
	merge_values(&(_clist->combine_items[_i].val), _clist->combine_items[_i].val, srbuffer[_i].val,
				 _info->type, _info->combine_strategy, _info->combineFn);
	return 0;
}

int merge_real_i(int * _base, int _one, int _two, int _combine_strategy, void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
	case ADD:
		*(_base) = _one + _two;
		break;
	case MULT:
		*(_base) = _one * _two;
		break;
	case AND:
		*(_base) = _one & _two;
		break;
	case OR:
		*(_base) = _one | _two;
		break;
	case FUNC:
		combineFn((void *)&_one, (void *)&_two, (void *)_base);
		break;
	}
	return 0;
}

int merge_real_f(float * _base, float _one, float _two, int _combine_strategy, void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
	case ADD:
		*(_base) = _one + _two;
		break;
	case MULT:
		*(_base) = _one * _two;
		break;
	//case AND:
	//	*(_base) = _one & _two;
	//	break;
	//case OR:
	//	*(_base) = _one | _two;
	//	break;
	case FUNC:
		combineFn((void *)&_one, (void *)&_two, (void *)_base);
		break;
	}
	return 0;
}

int merge_real_d(double * _base, double _one, double _two, int _combine_strategy, void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
	case ADD:
		*(_base) = _one + _two;
		break;
	case MULT:
		*(_base) = _one * _two;
		break;
	//case AND:
	//	*(_base) = _one & _two;
	//	break;
	//case OR:
	//	*(_base) = _one | _two;
	//	break;
	case FUNC:
		combineFn((void *)&_one, (void *)&_two, (void *)_base);
		break;
	}
	return 0;
}

int merge_values(Value * _dest, 
				 Value _val1, 
				 Value _val2, 
				 int _type, 
				 int _combine_strategy,
				 void (*combineFn)(void *, void *, void *))
{
	switch(_type) {
		case IVAR:
			merge_ints(_dest, _val1, _val2, _combine_strategy, combineFn);
			break;
		case FVAR:
			merge_floats(_dest, _val1, _val2, _combine_strategy, combineFn);
			break;
		case DVAR:
			merge_doubles(_dest, _val1, _val2, _combine_strategy, combineFn);
			break;
		default:
			;
	}

	return 0;
}

int merge_ints(Value * _dest, 
			   Value _val1, 
			   Value _val2,
			   int _combine_strategy,
			   void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
		case ADD:
			(*(_dest)).i = _val1.i + _val2.i;
			break;
		case MULT:
			(*(_dest)).i = _val1.i * _val2.i;
			break;
		case AND:
			(*(_dest)).i = _val1.i & _val2.i;
			break;
		case OR:
			(*(_dest)).i = _val1.i | _val2.i;
			break;
		case FUNC:
			combineFn(&_val1, &_val2, _dest);
			break;
		default:
			break;
	}

	return 0;
}

int merge_floats(Value * _dest, 
				 Value _val1, 
				 Value _val2,
				 int _combine_strategy,
				 void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
		case ADD:
			(*(_dest)).f = _val1.f + _val2.f;
			break;
		case MULT:
			(*(_dest)).f = _val1.f * _val2.f;
			break;
		//case AND:
		//	(*(_dest)).f = _val1.f & _val2.f;
		//	break;
		//case OR:
		//	(*(_dest)).f = _val1.f | _val2.f;
		//	break;
		case FUNC:
			combineFn(&_val1, &_val2, _dest);
			break;
		default:
			;
	}

	return 0;
}

int merge_doubles(Value * _dest, 
				  Value _val1, 
				  Value _val2,
				  int _combine_strategy,
				  void (*combineFn)(void *, void *, void *))
{
	switch(_combine_strategy) {
		case ADD:
			(*(_dest)).d = _val1.d + _val2.d;
			break;
		case MULT:
			(*(_dest)).d = _val1.d * _val2.d;
			break;
		//case AND:
		//	(*(_dest)).d = _val1.d & _val2.d;
		//	break;
		//case OR:
		//	(*(_dest)).d = _val1.d | _val2.d;
		//	break;
		case FUNC:
			combineFn(&_val1, &_val2, _dest);
			break;
		default:
			break;
	}

	return 0;
}

Value to_value(void * _base, int _type)
{
	Value val;
	switch(_type) {
		case IVAR:
			val.i = *((int *)_base);
			break;
		case FVAR:
			val.f = *((float *)_base);
			break;
		case DVAR:
			val.d = *((double *)_base);
			break;
		default:
			printf("Error in to_value\n");
			break;
	}
	return val;
}

int to_real_value(void * _base, int _type, Value _val)
{
	switch(_type) {
		case IVAR:
			*((int *)_base) = _val.i;
			break;
		case FVAR:
			*((float *)_base) = _val.f;
			break;
		case DVAR:
			*((double *)_base) = _val.d;
			break;
		default:
			printf("Error in to_real_value\n");
			break;
	}
	return 0;
}

/*
 * merge_rep_arrays
 * takes to arrays and merges them into one
 */
int merge_rep_arrays(void * _dest, void * _buffer, int _type, 
					 int _size, int _combine_strategy,
					 void (*_combineFn)(void *, void *, void *), void * hidden)
{
	switch(_type) {
		case IVAR:
			merge_rep_arrays_i((int *)_dest, (int *)_buffer, _size, 
				_combine_strategy, _combineFn, (int *)hidden);
			break;
		case FVAR:
			merge_rep_arrays_f((float *)_dest, (float *)_buffer, _size, 
				_combine_strategy, _combineFn, (float *)hidden);
			break;
		case DVAR:
			merge_rep_arrays_d((double *)_dest, (double *)_buffer, _size, 
				_combine_strategy, _combineFn, (double*) hidden);
			break;
		default:
			return MERGE_ERROR;
	}
	return MERGE_SUCCESS;
}

int merge_rep_arrays_i(int * _dest, int * _buffer, int _size, int _combine_strategy, 
					   void (*_combineFn)(void *, void *, void *), int * hidden)
{
	int i;
	switch(_combine_strategy) {
		case ADD:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] + _buffer[i];
			break;
		case MULT:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] * _buffer[i];
			break;
		case AND:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] & _buffer[i];
			break;
		case OR:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] | _buffer[i];
			break;
		case UPDATED:
			for(i = 0; i < _size; ++i)
				_dest[i] = (hidden[i] == _buffer[i]) ? _dest[i] : _buffer[i];
			break;
		case FUNC:
			_combineFn(&(_dest[i]), &(_buffer[i]), &(_dest[i]));
			break;
		default:
			return MERGE_ERROR;
	}
	return MERGE_SUCCESS;
}

int merge_rep_arrays_f(float * _dest, float * _buffer, int _size, int _combine_strategy,
				   void (*_combineFn)(void *, void *, void *), float * hidden)
{
	int i;
	switch(_combine_strategy) {
		case ADD:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] + _buffer[i];
			break;
		case MULT:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] * _buffer[i];
			break;
		//case AND:
		//	for(i = 0; i < _size; ++i)
		//		_dest[i] = _dest[i] & _buffer[i];
		//	break;
		//case OR:
		//	for(i = 0; i < _size; ++i)
		//		_dest[i] = _dest[i] | _buffer[i];
		//	break;
		case UPDATED:
			for(i = 0; i < _size; ++i)
				_dest[i] = (hidden[i] == _buffer[i]) ? _dest[i] : _buffer[i];
			break;
		case FUNC:
			_combineFn(&(_dest[i]), &(_buffer[i]), &(_dest[i]));
			break;
		default:
			return MERGE_ERROR;
	}
	return MERGE_SUCCESS;
}

int merge_rep_arrays_d(double * _dest, double * _buffer, int _size, int _combine_strategy,
				   void (*_combineFn)(void *, void *, void *), double * hidden)
{
	int i;
	switch(_combine_strategy) {
		case ADD:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] + _buffer[i];
			break;
		case MULT:
			for(i = 0; i < _size; ++i)
				_dest[i] = _dest[i] * _buffer[i];
			break;
		//case AND:
		//	for(i = 0; i < _size; ++i)
		//		_dest[i] = _dest[i] & _buffer[i];
		//	break;
		//case OR:
		//	for(i = 0; i < _size; ++i)
		//		_dest[i] = _dest[i] | _buffer[i];
		//	break;
		case UPDATED:
			for(i = 0; i < _size; ++i)
				_dest[i] = (hidden[i] == _buffer[i]) ? _dest[i] : _buffer[i];
			break;
		case FUNC:
			_combineFn(&(_dest[i]), &(_buffer[i]), &(_dest[i]));
			break;
		default:
			return MERGE_ERROR;
	}
	return MERGE_SUCCESS;
}

/*
 * array_size
 * returns SMALL, LARGE or ARRAY_SIZE_ERROR
 * used for determining if arrays should be sent for the UPDATED strategy
 */
int array_size(int _elements, int _type)
{
	int value;
	switch(_type) {
		case IVAR:
			value = (_elements > MAX_REP_ARRAY_INT) ? LARGE : SMALL;
			return value;
		case FVAR:
			value = (_elements > MAX_REP_ARRAY_FLOAT) ? LARGE : SMALL;
			return value;
		case DVAR:
			value = (_elements > MAX_REP_ARRAY_DOUBLE) ? LARGE : SMALL;
			return value;
		default:
			return ARRAY_SIZE_ERROR;
	}
}

/*
 * new_CombineList
 */
CombineList * new_CombineList(void)
{
	int i;
	CombineList * temp;
	
	// create space
	temp = (CombineList *)malloc(sizeof(CombineList));
	temp->combine_items = (CombineItem *)malloc(sizeof(CombineItem) * MAX_COMBINE_ITEMS);
	temp->combine_info	= (CombineInfo *)malloc(sizeof(CombineInfo) * MAX_COMBINE_ITEMS);
	temp->prefix_vars	= (PrefixVar *)malloc(sizeof(PrefixVar) * MAX_COMBINE_ITEMS);
	temp->rep_info		= (RepArrayInfo *)malloc(sizeof(RepArrayInfo) * (MAX_REPLICATED_ARRAYS + 1)); // +1 for N_DAmsgs
	temp->block_info	= (BlockArrayInfo *)malloc(sizeof(BlockArrayInfo) * MAX_BLOCK_ARRAYS);
	temp->cyclic_info	= (CyclicArrayInfo *)malloc(sizeof(CyclicArrayInfo) * MAX_CYCLIC_ARRAYS);
	temp->N_DAmsgs		= (int *)malloc(sizeof(int) * NestStep_get_size());

	// init variables
	temp->no_combine_items = 0;
	temp->no_prefix_vars = 0;
	temp->no_rep_arrays = 0;
	temp->largest_rep_array = 0;
	temp->no_block_arrays = 0;
	temp->no_cyclic_arrays = 0;
	temp->request_id_counter = 0;
	for(i = 0; i < NestStep_get_size(); ++i)
		temp->N_DAmsgs[i] = 0;

	return temp;
}

/*
 * free_CombineList
 */
void free_CombineList(CombineList * _clist)
{
	int i, j;
	for(i = 0; i < _clist->no_rep_arrays; ++i)
	{
		if(_clist->rep_info[i].prefix == PREFIX)
		{
			for(j = 0; j < thisgroup->tree->no_children; ++j)
				free(_clist->rep_info[i].child_arrays[j]);
			free(_clist->rep_info[i].child_arrays);
			free(_clist->rep_info[i].old_base);
		}
	}
	free(_clist->combine_items);
	free(_clist->combine_info);
	free(_clist->prefix_vars);
	free(_clist->rep_info);
	free(_clist->block_info);
	free(_clist->cyclic_info);
	free(_clist->N_DAmsgs);
	free(_clist);
}


/*
 * NestStep_read_request_b
 * issue a read request
 */
int NestStep_read_request_b(BlockDistArray * _array, 
							int _lower, int _higher, 
							void * _buffer, 
							CombineList * _clist)
{
	int i;
	int lower_limit;
	int higher_limit = -1;
	int elements;
	int ratio;
	int remainder;
	int element_size;
	ArrayRequest request;
	CombineList * combineList;

	if(_lower < 0 || _higher >= _array->size)
		return READ_REQUEST_ERROR;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;

	element_size = get_element_size(_array->type);

	ratio = _array->size / NestStep_get_size();
	remainder = _array->size % NestStep_get_size();
	
	request.name = _array->name;
	request.request = READ;
	request.lower = _lower;
	request.higher = _higher;
	request.request_id = combineList->request_id_counter;
	request.requester = NestStep_get_rank();
	
	// add to info
	combineList->block_info[combineList->no_block_arrays].block_array = _array;
	combineList->block_info[combineList->no_block_arrays].name = _array->name;
	combineList->block_info[combineList->no_block_arrays].buffer = _buffer;
	combineList->block_info[combineList->no_block_arrays].lower = _lower;
	combineList->block_info[combineList->no_block_arrays].higher = _higher;
	combineList->block_info[combineList->no_block_arrays].request = READ;
	combineList->block_info[combineList->no_block_arrays].request_id = combineList->request_id_counter;
	combineList->block_info[combineList->no_block_arrays].no_reads = 0;
	
	++combineList->request_id_counter;
	
	for(i = 0; i < NestStep_get_size(); ++i)
	{
		lower_limit = higher_limit + 1;
		elements = ratio + ((remainder > i) ? 1 : 0);
		higher_limit = lower_limit + elements - 1;

		if(	(lower_limit  >= _lower && lower_limit <= _higher)  ||
			(higher_limit >= _lower && higher_limit <= _higher) ||
			(lower_limit < _lower && higher_limit > _higher) )
		{
			if(i != NestStep_get_rank())
			{
				// increase N_DAmsgs
				++combineList->N_DAmsgs[i];
				
				// Send the request
				MPI_Bsend(&request, 
						1, 
						array_request_type, 
						i, 
						READ_REQUEST, 
						thisgroup->t_group.comm);

#ifdef DA_DEBUG
				printf("%d: sent request to %d: name: %d:%d range: %d:%d tag: %d\n", 
					NestStep_get_rank(),
					i,
					request.name.procedure, 
					request.name.relative,
					request.lower, request.higher,
					READ_REQUEST);
#endif
				++combineList->block_info[combineList->no_block_arrays].no_reads;
			}
			else
			{
				int start;
				int end;
				int amount;

				start = (_lower > _array->lower) ? _lower : _array->lower;
				end = (_higher < _array->higher) ? _higher : _array->higher;
				amount = end - start + 1;
				
				memcpy((char *)_buffer + (start - _lower) * element_size,
					   (char *)(_array->base) + (start - _array->lower) * element_size,
					   element_size * amount);
			}
		}
	}

	++combineList->no_block_arrays;

	return READ_REQUEST_SUCCESS;
}

/*
 * NestStep_write_request_b
 * issue a write request
 */
int NestStep_write_request_b(BlockDistArray * _array, 
							 int _lower, 
							 int _higher, 
							 void * _buffer, 
							 CombineList * _clist)
{
	int i;
	int lower_limit;
	int higher_limit = -1;
	int ratio;
	int remainder;
	int elements;
	int element_size;
	ArrayRequest request;
	MPI_Datatype type;
	CombineList * combineList;
#ifdef DA_DEBUG
	int k;
#endif
	
	if(_lower < 0 || _higher >= _array->size)
		return WRITE_REQUEST_ERROR;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;

	type = get_MPI_type(_array->type);
	element_size = get_element_size(_array->type);

	ratio = _array->size / NestStep_get_size();
	remainder = _array->size % NestStep_get_size();
	
	request.name = _array->name;
	request.request = WRITE;
	request.lower = _lower;
	request.higher = _higher;
	request.request_id = combineList->request_id_counter;
	request.requester = NestStep_get_rank();
	
	// add to info
	combineList->block_info[combineList->no_block_arrays].block_array = _array;
	combineList->block_info[combineList->no_block_arrays].name = _array->name;
	combineList->block_info[combineList->no_block_arrays].buffer = _buffer;
	combineList->block_info[combineList->no_block_arrays].lower = _lower;
	combineList->block_info[combineList->no_block_arrays].higher = _higher;
	combineList->block_info[combineList->no_block_arrays].request = WRITE;
	combineList->block_info[combineList->no_block_arrays].request_id = combineList->request_id_counter;
	combineList->block_info[combineList->no_block_arrays].no_reads = 0;
	
	++combineList->request_id_counter;

	for(i = 0; i < NestStep_get_size(); ++i)
	{
		lower_limit = higher_limit + 1;
		elements = ratio + ((remainder > i) ? 1 : 0);
		higher_limit = lower_limit + elements - 1;
		if(	(lower_limit  >= _lower && lower_limit <= _higher)  ||
			(higher_limit >= _lower && higher_limit <= _higher) ||
			(lower_limit < _lower && higher_limit > _higher) )
		{
			if(i != NestStep_get_rank())
			{
				int offset;
				int amount;
				offset = (lower_limit < _lower) ? 0 : (lower_limit - _lower);
				amount = ((higher_limit < _higher) ? higher_limit : _higher) - (_lower + offset) + 1;
#ifdef DA_DEBUG
				printf("%d: write request: %d lower: %d higher: %d offset: %d amount: %d\n",
					NestStep_get_rank(), i, lower_limit, higher_limit, offset, amount);
#endif
				// increase N_DAmsgs
				++combineList->N_DAmsgs[i];
				
				// Send the request
				MPI_Bsend(&request, 
						1, 
						array_request_type, 
						i, 
						WRITE_REQUEST, 
						thisgroup->t_group.comm);

				// send the elements
				MPI_Bsend((char *)_buffer + offset * element_size, amount, type, i,
						 REQUEST_ID_BASE + combineList->block_info[combineList->no_block_arrays].request_id,
						 thisgroup->t_group.comm);
#ifdef DA_DEBUG
				if(_array->type == IVAR)
				{
					printf("%d: sent write request to %d: %d:%d: elements:", 
						NestStep_get_rank(),
						i,
						request.lower,
						request.higher);
					for(k = 0; k < amount; ++k)
					{
						printf(" %d", *((char *)_buffer + (offset + k) * element_size));
					}
					printf("\n");
				}
#endif
			}
			else
			{
				int start;
				int end;
				int amount;

				start = (_lower > _array->lower) ? _lower : _array->lower;
				end = (_higher < _array->higher) ? _higher : _array->higher;
				amount = end - start + 1;
				
				memcpy((char *)_array->base + (start - _array->lower) * element_size,
					   (char *)_buffer + (start - _lower) * element_size,
					   element_size * amount);
			}
		}
	}

	++combineList->no_block_arrays;

	return WRITE_REQUEST_SUCCESS;
}

/*
 * NestStep_read_request_c
 */
int NestStep_read_request_c(CyclicDistArray * _array, int _lower, int _higher,
							void * _buffer, CombineList * _clist)
{
	int i, j;
	int start;
	int end;
	int element_size;
	int offset;
	ArrayRequest request;
	CombineList * combineList;

	if(_lower < 0 || _higher >= _array->size)
		return READ_REQUEST_ERROR;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;

	element_size = get_element_size(_array->type);

	request.name = _array->name;
	request.request = READ;
	request.lower = _lower;
	request.higher = _higher;
	request.request_id = combineList->request_id_counter;
	request.requester = NestStep_get_rank();
	
	// add to info
	combineList->cyclic_info[combineList->no_cyclic_arrays].cyclic_array = _array;
	combineList->cyclic_info[combineList->no_cyclic_arrays].name = _array->name;
	combineList->cyclic_info[combineList->no_cyclic_arrays].buffer = _buffer;
	combineList->cyclic_info[combineList->no_cyclic_arrays].lower = _lower;
	combineList->cyclic_info[combineList->no_cyclic_arrays].higher = _higher;
	combineList->cyclic_info[combineList->no_cyclic_arrays].request = READ;
	combineList->cyclic_info[combineList->no_cyclic_arrays].request_id = combineList->request_id_counter;
	combineList->cyclic_info[combineList->no_cyclic_arrays].no_reads = 0;

	++combineList->request_id_counter;

	for(i = 0; i < NestStep_get_size(); ++i)
	{
		for(j = 0;; ++j)
		{
			start = _array->blocksize * i +
				j * _array->blocksize * NestStep_get_size();
			end = start + _array->blocksize - 1;

			if(	(start  >= _lower && start <= _higher)  ||
				(end >= _lower && end <= _higher) ||
				(start < _lower && end > _higher) )
			{
				if(i == NestStep_get_rank())
				{
					// copy
					int offset;
					int first;
					int last;
					int amount;

					first = (_lower > start) ? _lower : start;
					last = (_higher > end) ? end : _higher;
					amount = last - first + 1;
					offset = first - _lower;

					memcpy((char *)_buffer + offset * element_size,
						   (char *)(_array->base) + global_local_index_c(_array, first) * element_size,
						   amount * element_size);
				}
				else
				{
					// add to N_DAmsgs, send request
					++combineList->N_DAmsgs[i];
					MPI_Bsend(&request, 
							 1, 
							 array_request_type, 
							 i, 
							 READ_REQUEST, 
							 thisgroup->t_group.comm);
#ifdef DA_DEBUG
					printf("%d: sent read request to %d: name: %d:%d limits: %d:%d\n",
						NestStep_get_rank(), i, 
						request.name.procedure,
						request.name.relative,
						request.lower,
						request.higher);
#endif

					++combineList->cyclic_info[combineList->no_cyclic_arrays].no_reads;
					break;
				}
			}
			else if(start > _higher || end > _higher)
			{
				break;
			}
		}
	}

	++combineList->no_cyclic_arrays;
	return READ_REQUEST_SUCCESS;
}

/*
 * NestStep_write_request_c
 */
int NestStep_write_request_c(CyclicDistArray * _array, int _lower, int _higher,
							 void * _buffer, CombineList * _clist)
{
	int i, j;
	MPI_Datatype type;
	int element_size;
	ArrayRequest request;
	int start;
	int end;
	int written;
	int read;
	int first;
	int last;
	int offset;
	int amount;
	void * buffer;
	int no_blocks;
	int max_blocks_node;
	CombineList * combineList;
#ifdef DA_DEBUG
	int k;
#endif

	if(_lower < 0 || _higher >= _array->size)
		return WRITE_REQUEST_ERROR;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;

	type = get_MPI_type(_array->type);
	element_size = get_element_size(_array->type);

	request.name = _array->name;
	request.request = WRITE;
	request.lower = _lower;
	request.higher = _higher;
	request.request_id = combineList->request_id_counter;
	request.requester = NestStep_get_rank();

	// add to info
	combineList->cyclic_info[combineList->no_cyclic_arrays].cyclic_array = _array;
	combineList->cyclic_info[combineList->no_cyclic_arrays].name = _array->name;
	combineList->cyclic_info[combineList->no_cyclic_arrays].buffer = _buffer;
	combineList->cyclic_info[combineList->no_cyclic_arrays].lower = _lower;
	combineList->cyclic_info[combineList->no_cyclic_arrays].higher = _higher;
	combineList->cyclic_info[combineList->no_cyclic_arrays].request = WRITE;
	combineList->cyclic_info[combineList->no_cyclic_arrays].request_id = combineList->request_id_counter;
	combineList->cyclic_info[combineList->no_cyclic_arrays].no_reads = 0;

	++combineList->request_id_counter;

	no_blocks = _array->size / _array->blocksize;
	max_blocks_node = (no_blocks / NestStep_get_size()) + 1;
	buffer = (void *)malloc(max_blocks_node * _array->blocksize * element_size);

	for(i = 0; i < NestStep_get_size(); ++i)
	{
		written = 0;
		read = 0;
		for(j = 0;; ++j)
		{
			// pack into buffer
			start = _array->blocksize * i +
				j * _array->blocksize * NestStep_get_size();
			end = start + _array->blocksize - 1;				
			if( (start  >= _lower && start <= _higher)  ||
				(end >= _lower && end <= _higher) ||
				(start < _lower && end > _higher) )
			{
				first = (_lower > start) ? _lower : start;
				last = (_higher > end) ? end : _higher;
				amount = last - first + 1;
				offset = first - _lower;

				memcpy((char *)buffer + written * element_size,
					(char *)_buffer + offset * element_size,
					amount * element_size);
				written += amount;
			}

			if(start >= _array->size ||end >= _array->size)
				break;
		}
		if(i == NestStep_get_rank())
		{
			// copy to array
			for(j = 0;; ++j)
			{
				start = _array->blocksize * i +
					j * _array->blocksize * NestStep_get_size();
				end = start + _array->blocksize - 1;				
				
				if( (start  >= _lower && start <= _higher)  ||
					(end >= _lower && end <= _higher) ||
					(start < _lower && end > _higher) )
				{
					first = (_lower > start) ? _lower : start;
					last = (_higher > end) ? end : _higher;
					amount = last - first + 1;

					memcpy((char *)_array->base + global_local_index_c(_array, first) * element_size,
						(char *)buffer + read * element_size,
						amount * element_size);
					read += amount;
				}

				if(start >= _array->size || end >= _array->size)
					break;
			}
		}
		else if(written > 0)
		{
			// increase N_DAmsgs
			++combineList->N_DAmsgs[i];
			
			// Send the request
			MPI_Bsend(&request, 
					 1, 
					 array_request_type, 
					 i, 
					 WRITE_REQUEST, 
					 thisgroup->t_group.comm);

			// send the elements
			MPI_Bsend((char *)buffer, written, type, i,
					 REQUEST_ID_BASE + combineList->cyclic_info[combineList->no_cyclic_arrays].request_id,
					 thisgroup->t_group.comm);

#ifdef DA_DEBUG
			if(_array->type == IVAR)
			{
				printf("%d: sent write request to %d: %d:%d: elements:", 
					NestStep_get_rank(),
					i,
					request.lower,
					request.higher);
				for(k = 0; k < written; ++k)
				{
					printf(" %d", ((int *)buffer)[k]);
				}
				printf("\n");
			}
#endif
		}
	}
	
	free(buffer);
	++_clist->no_cyclic_arrays;
	return WRITE_REQUEST_SUCCESS;
}

/*
 * process_DistArrays
 */
int process_DA_requests(CombineList * _clist, ArrayHolder * _holder)
{
	int i, j;
	MPI_Status status;
	ArrayRequest request;
	BlockDistArray * b_array;
	CyclicDistArray * c_array;
#ifdef DA_DEBUG
	int k, l;
	printf("%d:%d process_DA_requests: arrayholder b:", 
			NestStep_get_group_id(), NestStep_get_rank());
	for(l = 0; l < _holder->no_block_arrays; l++)
		printf(" %d:%d:", _holder->block_arrays[l]->name.procedure, _holder->block_arrays[l]->name.relative);
	printf("\n");

	printf("%d:%d process_DA_requests: arrayholder c:", 
			NestStep_get_group_id(), NestStep_get_rank());
	for(l = 0; l < _holder->no_cyclic_arrays; l++)
		printf(" %d:%d:", (*((_holder->cyclic_arrays)[l])).name.procedure, (*((_holder->cyclic_arrays)[l])).name.relative);
	printf("\n");
#endif
	// process requests
	for(i = 0; i < _clist->N_DAmsgs[NestStep_get_rank()]; ++i)
	{
		MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, thisgroup->t_group.comm, &status);
	
		if(status.MPI_TAG == READ_REQUEST)
		{
			MPI_Recv(&request, 1, array_request_type, status.MPI_SOURCE,
					status.MPI_TAG, thisgroup->t_group.comm, &status);
#ifdef DA_DEBUG
			printf("%d:%d process_DA_requests: looking for: %d:%d:", 
				NestStep_get_group_id(), NestStep_get_rank(),
				request.name.procedure, request.name.relative);
			for(k = 0; request.name.path[k] != -1; k++)
				printf("%d",request.name.path[k]);
			printf("\n");
#endif

			if((b_array = locate_array_b(_holder, request.name)) != NULL)
			{
				// process request b-array style
				int start;
				int end;
				int amount;
				int element_size;
				MPI_Datatype type;

#ifdef DA_DEBUG
				printf("%d:%d process_DA_requests: found in b: %d:%d:", 
					NestStep_get_group_id(), NestStep_get_rank(),
					request.name.procedure, request.name.relative);
				for(k = 0; request.name.path[k] != -1; k++)
					printf("%d",request.name.path[k]);
				printf("\n");
#endif

				start = (request.lower > b_array->lower) ? request.lower : b_array->lower;
				end = (request.higher < b_array->higher) ? request.higher : b_array->higher;
				amount = end - start + 1;

				type = get_MPI_type(b_array->type);
				element_size = get_element_size(b_array->type);

#ifdef DA_DEBUG
				if(b_array->type == IVAR)
				{
					printf("%d: process_DA_request: sending answer to %d:", 
						NestStep_get_rank(), request.requester);
					for(k = 0; k < amount; ++k)
					{
						printf(" %d",
							*((char *)b_array->base + (start - b_array->lower + k) * element_size));
					}
					printf("\n");
				}
#endif
				MPI_Bsend((char *)b_array->base + (start - b_array->lower) * element_size, 
						amount, type,
						request.requester, REQUEST_ID_BASE + request.request_id,
						thisgroup->t_group.comm);
			}
			else if((c_array = locate_array_c(_holder, request.name)) != NULL)
			{
				// process request c_array style
				int start;
				int end;
				int first;
				int last;
				int amount;
				int element_size;
				MPI_Datatype type;
				void * buffer;
				int no_blocks;
				int max_blocks_node;
				int written;

#ifdef DA_DEBUG
				printf("%d:%d process_DA_requests: found in c: %d:%d:", 
					NestStep_get_group_id(), NestStep_get_rank(),
					c_array->name.procedure, c_array->name.relative);
				for(k = 0; c_array->name.path[k] != -1; k++)
					printf("%d", c_array->name.path[k]);
				printf("\n");
#endif

				type = get_MPI_type(c_array->type);
				element_size = get_element_size(c_array->type);
				no_blocks = c_array->size / c_array->blocksize;
				max_blocks_node = (no_blocks / NestStep_get_size()) + 1;
				buffer = (void *)malloc(max_blocks_node * (c_array->blocksize) * element_size);
				
				// Pack data in buffer
				written = 0;
				for(j = 0;; ++j)
				{
					start = c_array->blocksize * NestStep_get_rank() +
						j * c_array->blocksize * NestStep_get_size();
					end = start + c_array->blocksize - 1;

#ifdef DA_DEBUG
					printf("%d:%d process_DA_requests: start == %d end == %d\n",
						NestStep_get_group_id(), NestStep_get_rank(),
						start, end);
#endif

					if(start >= c_array->size)
						break;
					if( (start  >= request.lower && start <= request.higher)  ||
						(end >= request.lower && end <= request.higher) ||
						(start < request.lower && end > request.higher) )
					{
						first = (request.lower > start) ? request.lower : start;
						last = (request.higher > end) ? end : request.higher;
						amount = last - first + 1;
						if(first > last)
							break;

#ifdef DA_DEBUG
						printf("%d: process_DA_requests: values: start: %d end: %d first: %d last: %d amount: %d local: %d\n",
							NestStep_get_rank(), start, end, first, last, amount, global_local_index_c(c_array, first));				
#endif
						memcpy(	(char *)buffer + written * element_size,
								(char *)c_array->base + global_local_index_c(c_array, first) * element_size,
								amount * element_size);

						written += amount;
					}

					if(end >= c_array->size)
						break;
				}
				
				MPI_Bsend(buffer, written, type, request.requester, 
					     REQUEST_ID_BASE + request.request_id,
						 thisgroup->t_group.comm);

#ifdef DA_DEBUG
				printf("%d: process_DA_requests: sending array to %d: elements:",
					NestStep_get_rank(),
					request.requester);
				for(k = 0; k < written; ++k)
				{
					if(c_array->type == IVAR)
						printf(" %d",((int *)buffer)[k]);
					else if(c_array->type == DVAR)
						printf(" %.2f", ((double *)buffer)[k]);
				}
				printf("\n");
#endif

				free(buffer);
			}
			else
			{
				; // return some error
			}
		}
		// process write requests
		else if(status.MPI_TAG == WRITE_REQUEST)
		{
			MPI_Recv(&request, 1, array_request_type, status.MPI_SOURCE,
					status.MPI_TAG, thisgroup->t_group.comm, &status);
			if((b_array = locate_array_b(_holder, request.name)) != NULL)
			{
				int start;
				int end;
				int amount;
				int element_size;
				MPI_Datatype type;

				start = (request.lower > b_array->lower) ? request.lower : b_array->lower;
				end = (request.higher < b_array->lower) ? request.higher : b_array->higher;
				amount = end  - start + 1;

				type = get_MPI_type(b_array->type);
				element_size = get_element_size(b_array->type);

				MPI_Recv((char *)b_array->base + (start - b_array->lower) * element_size, 
						amount, type, request.requester,
						REQUEST_ID_BASE + request.request_id,
						thisgroup->t_group.comm, &status);
			}
			else if((c_array = locate_array_c(_holder, request.name)) != NULL)
			{
				int start;
				int end;
				int first;
				int last;
				int read;
				int element_size;
				MPI_Datatype type;
				MPI_Status status;
				void * buffer;
				int count;
				int amount;

				type = get_MPI_type(c_array->type);
				element_size = get_element_size(c_array->type);
				
				MPI_Probe(request.requester, REQUEST_ID_BASE + request.request_id,
					thisgroup->t_group.comm, &status);
				MPI_Get_count(&status, type, &count);
				buffer = (void *)malloc(count * element_size);
				MPI_Recv(buffer, count, type, request.requester,
					 REQUEST_ID_BASE + request.request_id,
					 thisgroup->t_group.comm, &status);
				
				read = 0;
				for(j = 0;; ++j)
				{
					start = (c_array->blocksize) * NestStep_get_rank() +
						j * (c_array->blocksize) * NestStep_get_size();
					end = start + c_array->blocksize - 1;
					if( (start  >= request.lower && start <= request.higher)  ||
						(end >= request.lower && end <= request.higher) ||
						(start < request.lower && end > request.higher) )
					{
						if(start >= c_array->size)
							break;

						first = (request.lower > start) ? request.lower : start;
						last = (request.higher > end) ? end : request.higher;
						amount = last - first + 1;
						
						memcpy((char *)c_array->base + global_local_index_c(c_array, first) * element_size,
							(char *)buffer + read * element_size,
							amount * element_size);

						read += amount;
					}

					if(end >= c_array->size)
						break;
				}
				free(buffer);
			}
			else
			{
				; // do something useful
			}
		}
		else
		{
			; // error handling?
		}
	}
	return 0;
}

/*
 * process_DA_incoming
 * receives and processes messages from read requests
 */
int process_DA_incoming(CombineList * _clist, ArrayHolder * _holder)
{
	int i, j, k;
	int count;
	MPI_Status status;
	void * buffer;
	MPI_Datatype type;
	int element_size;
	int start_sender;
	int offset;
	int ratio;
	int remainder;

#ifdef DA_DEBUG
	int l;
#endif

	for(i = 0; i < _clist->no_block_arrays; ++i)
	{
		ratio = _clist->block_info[i].block_array->size / NestStep_get_size();
		remainder = _clist->block_info[i].block_array->size % NestStep_get_size();
		type = get_MPI_type(_clist->block_info[i].block_array->type);
		element_size = get_element_size(_clist->block_info[i].block_array->type);
		
		for(j = 0; j < _clist->block_info[i].no_reads; ++j)
		{
			// receive from any
			MPI_Probe(MPI_ANY_SOURCE, REQUEST_ID_BASE + _clist->block_info[i].request_id,
					  thisgroup->t_group.comm, &status);
			
			MPI_Get_count(&status, type, &count);

			start_sender = status.MPI_SOURCE * ratio + ((status.MPI_SOURCE < remainder) ? status.MPI_SOURCE : remainder);
			offset = (start_sender < _clist->block_info[i].lower) ? 0 : (start_sender - _clist->block_info[i].lower);

			MPI_Recv((char *)_clist->block_info[i].buffer + offset * element_size, 
					 count, type, status.MPI_SOURCE, 
					 REQUEST_ID_BASE + _clist->block_info[i].request_id,
					 thisgroup->t_group.comm, &status);
		}
	}

	for(i = 0; i < _clist->no_cyclic_arrays; ++i)
	{
		type = get_MPI_type(_clist->cyclic_info[i].cyclic_array->type);
		element_size = get_element_size(_clist->cyclic_info[i].cyclic_array->type);

		for(j = 0; j < _clist->cyclic_info[i].no_reads; ++j)
		{
			int amount;
			int read;
			int start, end;
			int first, last;

			MPI_Probe(MPI_ANY_SOURCE, REQUEST_ID_BASE + _clist->cyclic_info[i].request_id,
					  thisgroup->t_group.comm, &status);

			MPI_Get_count(&status, type, &count);

			buffer = (void *)malloc(count * element_size);

			MPI_Recv(buffer, count, type, status.MPI_SOURCE, 
					 REQUEST_ID_BASE + _clist->cyclic_info[i].request_id,
					 thisgroup->t_group.comm, &status);
			MPI_Get_count(&status, type, &count);
#ifdef DA_DEBUG
			printf("%d: process_DA_incoming: receiving array from %d: elements:",
				NestStep_get_rank(),
				status.MPI_SOURCE);
			for(l = 0; l < count; ++l)
			{
				if(_clist->cyclic_info[i].cyclic_array->type == IVAR)
					printf(" %d",((int *)buffer)[l]);
				else if(_clist->cyclic_info[i].cyclic_array->type == DVAR)
					printf(" %.2f", ((double *)buffer)[l]);
			}
			printf("\n");
#endif
			
			read = 0;
			for(k = 0;; ++k)
			{
				start = _clist->cyclic_info[i].cyclic_array->blocksize * status.MPI_SOURCE +
					k * _clist->cyclic_info[i].cyclic_array->blocksize * NestStep_get_size();
				end = start + _clist->cyclic_info[i].cyclic_array->blocksize - 1;

				if(start >= _clist->cyclic_info[i].cyclic_array->size)
					break;
				if( (start >= _clist->cyclic_info[i].lower && start <= _clist->cyclic_info[i].higher)  ||
					(end >= _clist->cyclic_info[i].lower && end <= _clist->cyclic_info[i].higher) ||
					(start < _clist->cyclic_info[i].lower && end > _clist->cyclic_info[i].higher) )
				{
					first = (_clist->cyclic_info[i].lower > start) ? _clist->cyclic_info[i].lower : start;
					last = (_clist->cyclic_info[i].higher > end) ? end : _clist->cyclic_info[i].higher;
					amount = last - first + 1;
					offset = first - _clist->cyclic_info[i].lower;

#ifdef DA_DEBUG
					printf("%d: process_DA_incoming: values: start: %d end: %d first: %d last: %d amount: %d\n",
						NestStep_get_rank(), start, end, first, last, amount);				
#endif

					memcpy( (char *)(_clist->cyclic_info[i].buffer) + 
								(first - _clist->cyclic_info[i].lower) * element_size,
							(char *)buffer + read * element_size,
							amount * element_size);
					read += amount;
				}

				if(end >= _clist->cyclic_info[i].cyclic_array->size)
					break;
			}

			free(buffer);
		}
	}
	return 0;
}

MPI_Datatype get_MPI_type(int _type)
{
	switch(_type) {
		case IVAR:
			return MPI_INT;
		case FVAR:
			return MPI_FLOAT;
		case DVAR:
			return MPI_DOUBLE;
	}
}

int get_element_size(int _type)
{
	switch(_type) {
		case IVAR:
			return sizeof(int);
		case FVAR:
			return sizeof(float);
		case DVAR:
			return sizeof(double);
	}

	return 0;
}
