#include "Combine_send_recv.h"
#include "NestStep.h"
#include "common.h"
#include <mpi.h>
#include <string.h>

CombineItem sitems[MAX_COMBINE_ITEMS];
int no_sitems = 0;
CombineItem ritems[MAX_COMBINE_ITEMS];
int no_ritems = 0;

RepArray * srarrays[MAX_REPLICATED_ARRAYS + 1];
int no_srarrays = 0;
RepArray * rrarrays[MAX_REPLICATED_ARRAYS + 1];
int no_rrarrays = 0;

int sdatacounter = 0;
int rdatacounter = 0;

char send_buffer[MAX_DATA];
char recv_buffer[MAX_DATA];

int * siptr = NULL;
int no_sints = 0;
float * sfptr = NULL;
int no_sfloats = 0;
double * sdptr = NULL;
int no_sdoubles = 0;
MPI_Datatype send_type;

int * riptr = NULL;
int no_rints = 0;
float * rfptr = NULL;
int no_rfloats = 0;
double * rdptr = NULL;
int no_rdoubles = 0;
MPI_Datatype recv_type;


int pack_send_CombineItem(CombineItem _item)
{
	sitems[no_sitems] = _item;
	no_sitems++;
	sdatacounter += sizeof(CombineItem);

	if(sdatacounter > RECOMMENDED_AMOUNT)
		return RECOMMENDED_AMOUNT_EXCEEDED;
	else
		return PACK_SUCCESS;
}

int pack_send_RepArray(RepArray * _array)
{
	srarrays[no_srarrays] = _array;
	no_srarrays++;
	sdatacounter += get_element_size(_array->type) * (_array->urh - _array->url + 1);

	if(sdatacounter > RECOMMENDED_AMOUNT)
		return RECOMMENDED_AMOUNT_EXCEEDED;
	else
		return PACK_SUCCESS;
}

int pack_send_buffer(void)
{
	int i = 0;
	char * temp = NULL;
	RepArray * cur_array = NULL;

	// combineitems
	for(i = 0; i < no_sitems; i++)
		((CombineItem *)send_buffer)[i] = sitems[i];
#ifdef SR_DEBUG
	for(i = 0; i < no_sitems; i++)
    {
		printf("%d: pack_send_buffer: item %d: name: %d:%d contributor: %d prefix_tag: %d\n",
			NestStep_get_rank(),
            i,
            ((CombineItem *)send_buffer)[i].name.procedure,
            ((CombineItem *)send_buffer)[i].name.relative,
            ((CombineItem *)send_buffer)[i].contributor,
            ((CombineItem *)send_buffer)[i].prefix_tag);
    }
#endif

	// doubles
	temp = &((CombineItem *)send_buffer)[no_sitems];
	if((((int)temp) % sizeof(double)) == 0)
		;
	else
		temp += sizeof(double) - (((int)temp) % sizeof(double));
	sdptr = (double *)temp;
	for(i = 0; i < no_srarrays; i++)
	{
		cur_array = srarrays[i];
		if(cur_array->type == DVAR)
		{
#ifdef SR_DEBUG
			printf("%d: pack_send_buffer: packing doubles: array: %d:%d size: %d no_sdoubles: %d\n",
			NestStep_get_rank(), 
			cur_array->name.procedure, 
			cur_array->name.relative,
			cur_array->size,
			no_sdoubles);
#endif
			memcpy(&sdptr[no_sdoubles], 
				(char *)(cur_array->base) + cur_array->url * get_element_size(cur_array->type), 
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			no_sdoubles += cur_array->urh - cur_array->url + 1;
		}
	}

	// floats
	temp = (char *)(&sdptr[no_sdoubles]);
	if((((int)temp) % sizeof(float)) == 0)
		;
	else
		temp += sizeof(float) - (((int)temp) % sizeof(float));
	sfptr = (float *)temp;
	for(i = 0; i < no_srarrays; i++)
	{
		cur_array = srarrays[i];
		if(cur_array->type == FVAR)
		{
			memcpy(&sfptr[no_sfloats], 
				(char *)(cur_array->base) + cur_array->url * get_element_size(cur_array->type), 
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			no_sfloats += cur_array->urh - cur_array->url + 1;
		}
	}

	// ints
	temp = (char *)(&sfptr[no_sfloats]);
	if((((int)temp) % sizeof(int)) == 0)
		;
	else
		temp += sizeof(int) - (((int)temp) % sizeof(int));
	siptr = (int *)temp;
	for(i = 0; i < no_srarrays; i++)
	{
		cur_array = srarrays[i];
		if(cur_array->type == IVAR)
		{
#ifdef SR_DEBUG
			printf("%d: pack_send_buffer: packing ints: array: %d:%d size: %d no_sints: %d\n",
			NestStep_get_rank(), 
			cur_array->name.procedure, 
			cur_array->name.relative,
			cur_array->size,
			no_sints);
#endif
			memcpy(&siptr[no_sints], 
				(char *)(cur_array->base) + cur_array->url * get_element_size(cur_array->type), 
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			no_sints += cur_array->urh - cur_array->url + 1;
		}
	}

	return PACK_SUCCESS;
}

int create_type_send(void)
{
	int i;
	MPI_Datatype vector_items;
	MPI_Datatype vector_doubles;
	MPI_Datatype vector_floats;
	MPI_Datatype vector_ints;

	int block_counts[4] = {1, 1, 1, 1};
	MPI_Datatype types[4];
	MPI_Aint displs[4];
	
	// create vector for combineitems
	MPI_Type_contiguous(no_sitems, combine_item_type, &vector_items);
	MPI_Type_commit(&vector_items);
	// create vector for doubles
	MPI_Type_contiguous(no_sdoubles, MPI_DOUBLE, &vector_doubles);
	MPI_Type_commit(&vector_doubles);
	// create vector for floats
	MPI_Type_contiguous(no_sfloats, MPI_FLOAT, &vector_floats);
	MPI_Type_commit(&vector_floats);
	// create vector for ints
	MPI_Type_contiguous(no_sints, MPI_INT, &vector_ints);
	MPI_Type_commit(&vector_ints);

	MPI_Address(send_buffer, &displs[0]);
	MPI_Address(sdptr, &displs[1]);
	MPI_Address(sfptr, &displs[2]);
	MPI_Address(siptr, &displs[3]);

	types[0] = vector_items;
	types[1] = vector_doubles;
	types[2] = vector_floats;
	types[3] = vector_ints;

	for(i = 3; i >= 0; i--)
		displs[i] -= displs[0];

	MPI_Type_struct(4, block_counts, displs, types, &send_type);
	MPI_Type_commit(&send_type);

#ifdef SR_DEBUG
	printf("%d: create_type_send: displs: %d %d %d %d no_sitems: %d no_sdoubles: %d no_sfloats: %d no_sints: %d\n",
		NestStep_get_rank(), displs[0], displs[1], displs[2], displs[3],
		no_sitems, no_sdoubles, no_sfloats, no_sints);
#endif

	
	MPI_Type_free(&vector_items);
	MPI_Type_free(&vector_doubles);
	MPI_Type_free(&vector_floats);
	MPI_Type_free(&vector_ints);

	return CREATE_TYPE_SUCCESS;
}

int send_data(int _dest, int _tag, MPI_Comm _comm)
{
	MPI_Send(send_buffer, 1, send_type, _dest, _tag, _comm);

	return 0;
}

int clear_send(void)
{
	MPI_Type_free(&send_type);
	no_sitems = 0;
	no_srarrays = 0;
	sdatacounter = 0;
	siptr = NULL;
	no_sints = 0;
	sfptr = NULL;
	no_sfloats = 0;
	sdptr = NULL;
	no_sdoubles = 0;

	return 0;
}

int req_recv_CombineItem(void)
{
	no_ritems++;
	rdatacounter += sizeof(CombineItem);

	if(rdatacounter > RECOMMENDED_AMOUNT)
		return RECOMMENDED_AMOUNT_EXCEEDED;
	else
		return REQ_SUCCESS;
}

int req_recv_RepArray(RepArray * _array)
{
	rrarrays[no_rrarrays] = _array;
	no_rrarrays++;

	rdatacounter += get_element_size(_array->type) * (_array->urh - _array->url + 1);

	if(rdatacounter > RECOMMENDED_AMOUNT)
		return RECOMMENDED_AMOUNT_EXCEEDED;
	else
		return REQ_SUCCESS;
}

int create_type_recv(void)
{
	int i;
	char * temp = NULL;
	RepArray * cur_array = NULL;
	MPI_Datatype vector_items;
	MPI_Datatype vector_doubles;
	MPI_Datatype vector_floats;
	MPI_Datatype vector_ints;

	int block_counts[4] = {1, 1, 1, 1};
	MPI_Datatype types[4];
	MPI_Aint displs[4];

	// find addresses for types
	// doubles
	temp = (char *)&((CombineItem *)recv_buffer)[no_ritems];
	if((((int)temp) % sizeof(double)) == 0)
		;
	else
		temp += sizeof(double) - (((int)temp) % sizeof(double));
	rdptr = (double *)temp;
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == DVAR)
			no_rdoubles += (cur_array->urh - cur_array->url + 1);
	}
	// floats
	temp = (char *)(&rdptr[no_rdoubles]);
	if((((int)temp) % sizeof(float)) == 0)
		;
	else
		temp += sizeof(float) - (((int)temp) % sizeof(float));
	rfptr = (float *)temp;
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == FVAR)
			no_rfloats += (cur_array->urh - cur_array->url + 1);
	}
	// ints
	temp = (char *)(&rfptr[no_rfloats]);
	if((((int)temp) % sizeof(int)) == 0)
		;
	else
		temp += sizeof(int) - (((int)temp) % sizeof(int));
	riptr = (int *)temp;
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == IVAR)
			no_rints += (cur_array->urh - cur_array->url + 1);
	}
	
	
	// create vector for combineitems
	MPI_Type_contiguous(no_ritems, combine_item_type, &vector_items);
	MPI_Type_commit(&vector_items);
	// create vector for doubles
	MPI_Type_contiguous(no_rdoubles, MPI_DOUBLE, &vector_doubles);
	MPI_Type_commit(&vector_doubles);
	// create vector for floats
	MPI_Type_contiguous(no_rfloats, MPI_FLOAT, &vector_floats);
	MPI_Type_commit(&vector_floats);
	// create vector for ints
	MPI_Type_contiguous(no_rints, MPI_INT, &vector_ints);
	MPI_Type_commit(&vector_ints);

	MPI_Address(recv_buffer, &displs[0]);
	MPI_Address(rdptr, &displs[1]);
	MPI_Address(rfptr, &displs[2]);
	MPI_Address(riptr, &displs[3]);

	types[0] = vector_items;
	types[1] = vector_doubles;
	types[2] = vector_floats;
	types[3] = vector_ints;

	for(i = 3; i >= 0; i--)
		displs[i] -= displs[0];

	MPI_Type_struct(4, block_counts, displs, types, &recv_type);
	MPI_Type_commit(&recv_type);

#ifdef SR_DEBUG
	printf("%d: create_type_recv: displs: %d %d %d %d no_ritems: %d no_rdoubles: %d no_rfloats: %d no_rints: %d\n",
		NestStep_get_rank(), displs[0], displs[1], displs[2], displs[3],
		no_ritems, no_rdoubles, no_rfloats, no_rints);
#endif
	
	MPI_Type_free(&vector_items);
	MPI_Type_free(&vector_doubles);
	MPI_Type_free(&vector_floats);
	MPI_Type_free(&vector_ints);

	return CREATE_TYPE_SUCCESS;
}

int recv_data(int _source, int _tag, MPI_Comm _comm, MPI_Status * _status)
{
	MPI_Recv(recv_buffer, 1, recv_type, _source, _tag, _comm, _status);

	return 0;
}
int unpack_recv_buffer(void)
{
	int i;
	int read_doubles = 0;
	int read_floats = 0;
	int read_ints = 0;
	RepArray * cur_array;
	// doubles
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == DVAR)
		{
			memcpy((char *)cur_array->base + cur_array->url * get_element_size(cur_array->type),
				&rdptr[read_doubles],
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			read_doubles += (cur_array->urh - cur_array->url + 1);
		}
	}
	// floats
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == FVAR)
		{
			memcpy((char *)cur_array->base + cur_array->url * get_element_size(cur_array->type),
				&rfptr[read_floats],
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			read_floats += (cur_array->urh - cur_array->url + 1);
		}
	}
	// ints
	for(i = 0; i < no_rrarrays; i++)
	{
		cur_array = rrarrays[i];
		if(cur_array->type == IVAR)
		{
			memcpy((char *)cur_array->base + cur_array->url * get_element_size(cur_array->type),
				&riptr[read_ints],
				(cur_array->urh - cur_array->url + 1) * get_element_size(cur_array->type));
			read_ints += (cur_array->urh - cur_array->url + 1);
		}
	}

	return UNPACK_SUCCESS;
}

void * get_rep_start(RepArray * _array)
{
	int i;
	int elements = 0;
	int element_size;
	char * pointer;
	
	element_size = get_element_size(_array->type);
	switch(_array->type) {
		case IVAR:
			pointer = (char *)riptr;
			break;
		case FVAR:
			pointer = (char *)rfptr;
			break;
		case DVAR:
			pointer = (char *)rdptr;
			break;
	}
	for(i = 0; i < no_rrarrays; i++)
	{
		if(compare_name(_array->name, rrarrays[i]->name) == 0)
			break;
		else if(_array->type == rrarrays[i]->type)
			elements += rrarrays[i]->urh - rrarrays[i]->url + 1;
	}
	
	return (void *)(pointer + elements * element_size);
}

int clear_recv(void)
{
	MPI_Type_free(&recv_type);
	no_ritems = 0;
	no_rrarrays = 0;
	rdatacounter = 0;
	riptr = NULL;
	no_rints = 0;
	rfptr = NULL;
	no_rfloats = 0;
	rdptr = NULL;
	no_rdoubles = 0;

	return 0;
}
