#include "NestStep.h"
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <string.h>
#include "tree_builders.h"

Group * root;
Group * thisgroup;
MPI_Datatype combine_item_type;
MPI_Datatype array_request_type;
void * bbuffer;
ArrayHolder * arrayHolder;
CombineList * combineList;

void NestStep_init(int * pargc, char *** pargv)
{
	int i;
	// variable declarations for datatype creation

	// for combine items
	int blockcounts_item[4] = {1, 1, 2, 1};
	MPI_Datatype types_item[4];
	MPI_Aint displs_item[4];
	CombineItem item[2];

	MPI_Datatype valuetype;
	int blockcounts_val[2] = {1, 1};
	MPI_Datatype types_val[2];
	MPI_Aint displs_val[2];
	Value val[2];

	MPI_Datatype nametype;
	int blockcounts_name[2] = {2 + MAX_DEPTH, 1};
	MPI_Datatype types_name[2];
	MPI_Aint displs_name[2];
	Name name[2];

	// for array requests
	int blockcounts_request[3] = {1, 5, 1};
	MPI_Datatype types_request[3];
	MPI_Aint displs_request[3];
	ArrayRequest request[2];
	
	/* initialize root */
	root = new_Group();	
	/* initiate tlib */
	T_Init(pargc, pargv, MPI_COMM_WORLD, &root->t_group);

	// initiate root group
	thisgroup = root;
	thisgroup->clist = new_CombineList();
	new_Tree();

	// initiate path
	thisgroup->path[0] = 0;
	thisgroup->path[1] = -1;

	// create the mpi datatype(s)
	// valuetype
	MPI_Address(&val[0].d, &displs_val[0]);
	MPI_Address(&val[1].d, &displs_val[1]);
	types_val[0] = MPI_DOUBLE;
	types_val[1] = MPI_UB;
	for(i = 1; i >= 0; --i)
		displs_val[i] -= displs_val[0];
	MPI_Type_struct(2, blockcounts_val, displs_val, types_val, &valuetype);
	MPI_Type_commit(&valuetype);

	// nametype
	MPI_Address(&name[0].procedure, &displs_name[0]);
	MPI_Address(&name[1].procedure, &displs_name[1]);
	types_name[0] = MPI_INT;
	types_name[1] = MPI_UB;
	for(i = 1; i >= 0; --i)
		displs_name[i] -= displs_name[0];
	MPI_Type_struct(2, blockcounts_name, displs_name, types_name, &nametype);
	MPI_Type_commit(&nametype);

	// itemtype
	MPI_Address(&item[0].name,			&displs_item[0]);
	MPI_Address(&item[0].val,			&displs_item[1]);
	MPI_Address(&item[0].contributor,	&displs_item[2]);
	MPI_Address(&item[1].name,			&displs_item[3]);
	types_item[0] = nametype;
	types_item[1] = valuetype;
	types_item[2] = MPI_INT;
	types_item[3] = MPI_UB;
	for(i = 3; i >= 0; --i)
		displs_item[i] -= displs_item[0];
	MPI_Type_struct(4, blockcounts_item, displs_item, types_item, &combine_item_type);
	MPI_Type_commit(&combine_item_type);


	// array_request_type
	MPI_Address(&request[0].name,		&displs_request[0]);
	MPI_Address(&request[0].request,	&displs_request[1]);
	MPI_Address(&request[1].name,		&displs_request[2]);
	types_request[0] = nametype;
	types_request[1] = MPI_INT;
	types_request[2] = MPI_UB;
	for(i = 2; i >= 0; --i)
		displs_request[i] -= displs_request[0];
	MPI_Type_struct(3, blockcounts_request, displs_request, types_request, &array_request_type);
	MPI_Type_commit(&array_request_type);

	// free temporary datatypes
	MPI_Type_free(&nametype);
	MPI_Type_free(&valuetype);

	/////////////////////////
	// Initialize send buffer
	/////////////////////////
	bbuffer = (void *)malloc(BBUFFERSIZE);
	MPI_Buffer_attach(bbuffer, BBUFFERSIZE);
	arrayHolder = new_ArrayHolder();
}

void NestStep_finalize(void)
{
	int size;

	// free datatype(s)
	MPI_Type_free(&combine_item_type);
	MPI_Type_free(&array_request_type);

	/* free root */
	free_CombineList(thisgroup->clist);
	free_Tree();
	free(root);

	free_ArrayHolder(arrayHolder);

	// free send buffer
	MPI_Buffer_detach(&bbuffer, &size);
	free(bbuffer);

	/* close down tlib */
	T_Finalize();
}

int NestStep_get_rank(void)
{
	return T_GetMyRankLocal(&thisgroup->t_group);
}

int NestStep_get_size(void)
{
	return T_GetMyGroupSize(&thisgroup->t_group);
}

int NestStep_get_group_id(void)
{
	return T_GetMyGroupNo(&thisgroup->t_group);
}

Group * NestStep_get_parent_group(void)
{
	return thisgroup->parent;
}

void NestStep_step(void)
{
	int i;
	for(i = 0; thisgroup->path[i] != -1; i++)
		;
	thisgroup->path[i] = NestStep_get_group_id();
	thisgroup->path[i + 1] = -1;
	// thisgroup->clist = new_CombineList();
}

void NestStep_step_end(void)
{
	int i;
	for(i = 0; thisgroup->path[i] != -1; i++)
		;
	thisgroup->path[i - 1] = -1;
	// free_CombineList(thisgroup->clist);
}

void NestStep_neststep(int k)
{
	/* declarations */
	int i;
	float * weight;
	
	/* init */
/*
	weight = (float *)malloc(k * sizeof(float));
	for(i = 0; i < k; ++i)
		weight[i] = 1.0 / (float)k;

	NestStep_neststep_w(k, weight);
	free(weight);
	*/
	NestStep_neststep_c(k, NestStep_get_rank() % k);
}

void NestStep_neststep_w(int k, float weight[])
{
	/* declarations */
	int i;
	Group * temp_group;
	int result;
	
	/* code */

	temp_group = new_Group();
	
	result = T_Split_GrpParfor(k, &thisgroup->t_group, &temp_group->t_group, weight);
	if(result != 0)
	{
		/* do some better error handling here */
		printf("Error creating tlib groups!\n");
	}
	// copy old path
	memcpy(&(temp_group->path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));
	/* set thisgroup to the newly created group */
	thisgroup = temp_group;
	// add to path
	for(i = 0; thisgroup->path[i] != -1; i++)
		;
	thisgroup->path[i] = NestStep_get_group_id();
	thisgroup->path[i + 1] = -1;

	thisgroup->clist = new_CombineList();
	new_Tree();
}

void NestStep_neststep_c(int k, int color)
{
	/* declarations */
	Group * temp_group;
	int result;

	/* code */

	temp_group = new_Group();
	
	/* in order that no wierdness shows up, execute a split for alla members of the group and let the ones */
	/* that should ignore the statement end the step at once */
	if(color < k && color >=0)
		result = T_Split_GrpExpl(&thisgroup->t_group, &temp_group->t_group, color, 0); /* last zero is for key */
	else
		result = T_Split_GrpExpl(&thisgroup->t_group, &temp_group->t_group, k, 0); /* last zero is for key */
	if(result != 0)
	{
		/* do some better error handling here */
		printf("Error creating tlib groups!\n");
	}

	if(color < k && color >=0)
	{
		int i;
		// copy old path
		memcpy(&(temp_group->path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));
		/* set thisgroup to the newly created group */
		thisgroup = temp_group;
		// add to path
		for(i = 0; thisgroup->path[i] != -1; i++)
			;
		thisgroup->path[i] = NestStep_get_group_id();
		thisgroup->path[i + 1] = -1;
		thisgroup->clist = new_CombineList();
		new_Tree();
	}
	else
	{
		/* this means the statement should be ignored */
		/* free temp_group and continue in original group, no combine or anything */
		free_Group(temp_group);
	}
}

void NestStep_neststep_end(void)
{
	/* declarations */
	Group * temp_group;

	// free up some stuff
	free_Tree();
	free_CombineList(thisgroup->clist);

	/* return to parent and free child */
	temp_group = thisgroup;
	thisgroup = thisgroup->parent;
	free_Group(temp_group);
}

/*
 * new_Group
 * creates a new group and initializes most things
 * does not initialize t_group, tree and N_DAmsgs
 */
Group * new_Group(void)
{
	Group * temp_group;

	temp_group = (Group *)malloc(sizeof(Group));
	temp_group->parent = thisgroup;
	
	return temp_group;
}

void free_Group(Group * group)
{
	free(group);
}

int new_Tree(void)
{
	int i;

	thisgroup->tree = (Tree *)malloc(sizeof(Tree));
	thisgroup->tree->children = (int *)malloc((NestStep_get_size() - 1) * sizeof(int));
	
	build_tree(NestStep_get_size(),
			   NestStep_get_rank(),
			   &(thisgroup->tree->parent),
			   &(thisgroup->tree->no_children),
			   &(thisgroup->tree->children));
#ifdef TREE_DEBUG
	printf("%d:%d: new_Tree: size of group: %d parent: %d children:",
		NestStep_get_group_id(),
		NestStep_get_rank(),
		NestStep_get_size(),
		thisgroup->tree->parent);
	for(i = 0; i < thisgroup->tree->no_children; i++)
		printf(" %d", thisgroup->tree->children[i]);
	printf("\n");
#endif

	return 0;
}

int free_Tree(void)
{
	free(thisgroup->tree->children);
	free(thisgroup->tree);

	return 0;
}

/*
 * compare_name
 * returns 0 if the names are equal
 */
int compare_name(Name name1, Name name2)
{
	if(name1.procedure < name2.procedure)
		return -1;
	else if(name1.procedure > name2.procedure)
		return 1;
	else if(name1.relative < name2.relative)
		return -1;
	else if(name1.relative > name2.relative)
		return 1;
	else
	{
		int i;
		for(i = 0; name1.path[i] != -1 && name2.path[i] != -1; i++)
		{
			if(name1.path[i] < name2.path[i])
				return -1;
			else if(name1.path[i] > name2.path[i])
				return 1;
		}
		return 0;
	}
}

