#include "ShVar.h"
#include "NestStep.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/*
 * new_ShVar
 * creates a new shared variable
 */
ShVar * new_ShVar(Name name, int type, void * base)
{
	ShVar * tempvar = (ShVar *)malloc(sizeof(ShVar));
	tempvar->name.procedure = name.procedure;
	tempvar->name.relative = name.relative;
	tempvar->type = type;
	tempvar->base = base;
	memcpy(&(tempvar->name.path[0]), &(thisgroup->path[0]), MAX_DEPTH * sizeof(int));

	if(type == IVAR && base != NULL)
	{
		tempvar->hidden = (void *)malloc(sizeof(int));
		memcpy(tempvar->hidden, base, sizeof(int));
		
	}
	else if(type == FVAR && base != NULL)
	{
		tempvar->hidden = (void *)malloc(sizeof(float));
		memcpy(tempvar->hidden, base, sizeof(float));
	}
	else if(type == DVAR && base != NULL)
	{
		tempvar->hidden = (void *)malloc(sizeof(double));
		memcpy(tempvar->hidden, base, sizeof(double));
	}
	else
		tempvar->hidden = NULL;

	return tempvar;
}

/*
 * free_ShVar
 * frees up the memory used by a shared variable
 */
void free_ShVar(ShVar * var)
{
	if(var->hidden != NULL)
		free(var->hidden);
	free(var);
}

/*
 * NestStep_shvar_attach
 * returns ATTACH_SUCCESS if the variable was successfully inserted in the container
 * returns ATTACH_ERROR otherwise
 */
int NestStep_shvar_attach(ShVar * var, 
						  int combine_strategy, 
						  void (*combineFn)(void *, void *, void *), 
						  int prefix_tag,
						  void * prefix_base,
						  int _updated,
						  CombineList * _clist)
{
	CombineList * combineList;
	if(_clist == NULL)
		combineList = thisgroup->clist;
	else
		combineList = _clist;
	if(combineList->no_combine_items < MAX_COMBINE_ITEMS)
	{
		/* create combineinfo item */
		combineList->combine_info[combineList->no_combine_items].var = var;
		combineList->combine_info[combineList->no_combine_items].combine_strategy = (combine_strategy == LEADER) ? UPDATED : combine_strategy;
		combineList->combine_info[combineList->no_combine_items].combineFn = combineFn;
		combineList->combine_info[combineList->no_combine_items].type = var->type;
		if(combine_strategy == LEADER && NestStep_get_rank() == 0)
			combineList->combine_info[combineList->no_combine_items].updated = UPDATED;
		else if(combine_strategy == LEADER && NestStep_get_rank() != 0)
			combineList->combine_info[combineList->no_combine_items].updated = NOT_UPDATED;
		else
			combineList->combine_info[combineList->no_combine_items].updated = _updated;
		
		/* create combineitem */
		combineList->combine_items[combineList->no_combine_items].name = var->name;		
		combineList->combine_items[combineList->no_combine_items].contributor = NestStep_get_rank();

		if(combine_strategy == LEADER && NestStep_get_rank() == 0)
		{
			switch(var->type) {
				case IVAR:
					memset(var->hidden, 0, sizeof(int));
					break;
				case FVAR:
					memset(var->hidden, 0, sizeof(float));
					break;
				case DVAR:
					memset(var->hidden, 0, sizeof(double));
					break;
				default:
					return ATTACH_ERROR;
			}
		}
		else if(combine_strategy == LEADER && NestStep_get_rank() != 0)
		{
			memset(&(combineList->combine_items[combineList->no_combine_items].val), 0, sizeof(Value));
			//combineList->combine_items[combineList->no_combine_items].prefix_tag = NOT_UPDATED;
		}
		else
			combineList->combine_items[combineList->no_combine_items].prefix_tag = (_updated == UPDATED) ? _updated : prefix_tag;
		
		switch(var->type) {
			case IVAR:
				combineList->combine_items[combineList->no_combine_items].val.i = *((int *)var->base);
				break;
			case FVAR:
				combineList->combine_items[combineList->no_combine_items].val.f = *((float *)var->base);
				break;
			case DVAR:
				combineList->combine_items[combineList->no_combine_items].val.d = *((double *)var->base);
				break;
			default:
				return ATTACH_ERROR;
		}

		/* create prefix variable if necessary */
		if(prefix_tag == PREFIX)
		{
			combineList->prefix_vars[combineList->no_prefix_vars].prefix = prefix_base;
			combineList->prefix_vars[combineList->no_prefix_vars].shared = var;
			combineList->prefix_vars[combineList->no_prefix_vars].child_values = 
				(Value *)malloc(thisgroup->tree->no_children * sizeof(Value));
			switch(var->type) {
				case IVAR:
					combineList->prefix_vars[combineList->no_prefix_vars].my_value.i = *((int *)var->base);
					break;
				case FVAR:
					combineList->prefix_vars[combineList->no_prefix_vars].my_value.f = *((float *)var->base);
					break;
				case DVAR:
					combineList->prefix_vars[combineList->no_prefix_vars].my_value.d = *((double *)var->base);
					break;
				default:
					return ATTACH_ERROR;
			}
			++combineList->no_prefix_vars;
			qsort((void *)combineList->prefix_vars,
				  combineList->no_prefix_vars,
				  sizeof(PrefixVar),
				  compare_prefix);
		}

		++combineList->no_combine_items;
		qsort((void *)combineList->combine_info,
			  combineList->no_combine_items,
			  sizeof(CombineInfo),
			  compare_info);
		return ATTACH_SUCCESS;
	}
	else
		return ATTACH_ERROR;
}

void set_zero(ShVar * _var)
{
	switch(_var->type) {
		case IVAR:
			*((int *)_var->base) = 0;
			break;
		case FVAR:
			*((float *)_var->base) = 0.0;
			break;
		case DVAR:
			*((double *)_var->base) = 0.0;
			break;
	}
}
