#include "NestStep.h"
#include "import_export.h"
#include "arrays.h"
#include <string.h>

int importArray_rep_to_rep(RepArray * _from, RepArray * _to,
						   int _lower, int _higher, int active)
{
	int element_size;
	int amount;

	if(active == 0)
		return importArray_rep_to_rep(_from, _to, -1, -1, 1);

	if(_from->type == _to->type)
	{
		element_size = get_element_size(_to->type);
		amount = _higher - _lower + 1;
		memcpy(_to->base, (char *)_from->base + _lower * element_size,
			amount * element_size);
		return IMPORT_SUCCESS;
	}
	else
		return IMPORT_ERROR;
}

int exportArray_rep_to_rep(RepArray * _from, RepArray * _to,
						   int _lower, int _higher, int active)
{
	int element_size;
	int amount;

	if(active == 0)
		return exportArray_rep_to_rep(_from, _to, -1, -1, 1);

	if(_from->type == _to->type)
	{
		element_size = get_element_size(_to->type);
		amount = _higher - _lower + 1;
		//nolla
		memcpy((char *)_to->base + _lower * element_size, _from->base,
			amount * element_size);
		// combine med add
		return EXPORT_SUCCESS;
	}
	else
		return EXPORT_ERROR;
}

/*
 * importArray_b_to_b
 * _from must be created in parent group and _to in target group
 * assumes that the requested amount match the size of the array
 */
int importArray_b_to_b(BlockDistArray * _from, BlockDistArray * _to,
					   int _lower, int _higher, int active)
{
	int lower;
	int higher;
	CombineList * clist;
	ArrayHolder * holder;
	Group * holder_cur_group;

	if(active == 0)
		return importArray_b_to_b(_from, _to, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;

	clist = new_CombineList();
	holder = new_ArrayHolder();

	lower = _lower + _to->lower;
	higher = lower + _to->higher - _to->lower;
	
	NestStep_register_b(holder, _from);

	if(_lower != -1 && _higher != -1)
		NestStep_read_request_b(_from, lower, higher,
			_to->base, clist);

	NestStep_combine(clist, holder);
	
	free_CombineList(clist);
	free_ArrayHolder(holder);

	thisgroup = holder_cur_group;

	return IMPORT_SUCCESS;
}

/*
 * exportArray_b_to_b
 * _from must be created in child group and _to in parent group
 * assumes that the requested amount match the size of the array
 */
int exportArray_b_to_b(BlockDistArray * _from, BlockDistArray * _to,
					   int _lower, int _higher, int active)
{
	int lower;
	int higher;
	CombineList * clist;
	ArrayHolder * holder;
	Group * holder_cur_group;

	if(active == 0)
		return exportArray_b_to_b(_from, _to, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;

	clist = new_CombineList();
	holder = new_ArrayHolder();

	lower = _lower + _from->lower;
	higher = lower + _from->higher - _from->lower;

	NestStep_register_b(holder, _to);

	if(_lower != -1 && _higher != -1)
		NestStep_write_request_b(_to, lower, higher,
			_from->base, clist);

	NestStep_combine(clist, holder);
	
	free_CombineList(clist);
	free_ArrayHolder(holder);

	thisgroup = holder_cur_group;

	return EXPORT_SUCCESS;
}

int importArray_b_to_rep(BlockDistArray * _barray, RepArray * _rarray,
						 int _lower, int _higher, int active)
{
	Group * holder_cur_group;
	CombineList * clist;
	ArrayHolder * holder;

	if(active == 0)
		return importArray_b_to_rep(_barray, _rarray, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;
	clist = new_CombineList();
	holder = new_ArrayHolder();

	NestStep_register_b(holder, _barray);

	if(_lower == -1 && _higher == -1)
		NestStep_read_request_b(_barray, _lower, _higher,
			_rarray->base, clist);

	NestStep_combine(clist, holder);

	thisgroup = holder_cur_group;
	free_CombineList(clist);
	free_ArrayHolder(holder);

	return IMPORT_SUCCESS;
}

int exportArray_rep_to_b(RepArray * _rarray, BlockDistArray * _barray,
						 int _lower, int _higher, int active)
{
	Group * holder_cur_group;
	CombineList * clist;
	ArrayHolder * holder;

	if(active == 0)
		return exportArray_rep_to_b(_rarray, _barray, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;
	clist = new_CombineList();
	holder = new_ArrayHolder();

	NestStep_register_b(holder, _barray);

	if(_lower == -1 && _higher == -1)
		NestStep_write_request_b(_barray, _lower, _higher,
			_rarray->base, clist);

	NestStep_combine(clist, holder);

	thisgroup = holder_cur_group;
	free_CombineList(clist);
	free_ArrayHolder(holder);

	return EXPORT_SUCCESS;
}

int importArray_c_to_rep(CyclicDistArray * _carray, RepArray * _rarray,
						 int _lower, int _higher, int active)
{
	Group * holder_cur_group;
	CombineList * clist;
	ArrayHolder * holder;

	if(active == 0)
		return importArray_c_to_rep(_carray, _rarray, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;
	clist = new_CombineList();
	holder = new_ArrayHolder();

	NestStep_register_c(holder, _carray);

	if(_lower == -1 && _higher == -1)
		NestStep_read_request_c(_carray, _lower, _higher,
			_rarray->base, clist);

	NestStep_combine(clist, holder);

	thisgroup = holder_cur_group;
	free_CombineList(clist);
	free_ArrayHolder(holder);

	return IMPORT_SUCCESS;
}

int exportArray_rep_to_c(RepArray * _rarray, CyclicDistArray * _carray,
						 int _lower, int _higher, int active)
{
	Group * holder_cur_group;
	CombineList * clist;
	ArrayHolder * holder;

	if(active == 0)
		return exportArray_rep_to_c(_rarray, _carray, -1, -1, 1);

	holder_cur_group = thisgroup;
	thisgroup = thisgroup->parent;
	clist = new_CombineList();
	holder = new_ArrayHolder();

	NestStep_register_c(holder, _carray);

	if(_lower == -1 && _higher == -1)
		NestStep_write_request_c(_carray, _lower, _higher,
			_rarray->base, clist);

	NestStep_combine(clist, holder);

	thisgroup = holder_cur_group;
	free_CombineList(clist);
	free_ArrayHolder(holder);

	return EXPORT_SUCCESS;
}
