#ifndef __ARRAYS
#define __ARRAYS

#include "common.h"

#define READ 1
#define WRITE 2
#define MAX_REPLICATED_ARRAYS 64
#define MAX_BLOCK_ARRAYS 64
#define MAX_CYCLIC_ARRAYS 64

typedef struct reparray {
	Name name;
	int type;				// IVAR, FVAR, DVAR
	int size;				// no of elements in the array
	void * base;
	void * hidden;
	int url;				// update range lower
	int urh;				// update range higher
} RepArray;

typedef struct cyclicdistarray {
	Name name;
	int type;				// IVAR, FVAR, DVAR
	int size;				// number of elements in the array
	int blocksize;			// size of the smallest parts
	void * base;
} CyclicDistArray;

typedef struct blockdistarray {
	Name name;
	int type;				// IVAR, FVAR, DVAR
	int size;				// number of elements in the array
	int local_size;
	int lower;
	int higher;
	void * base;
} BlockDistArray;

typedef struct arrayrequest {
	Name name;
	int request;			// READ or WRITE
	int lower;
	int higher;
	int request_id;
	int requester;			// is that even a word?
} ArrayRequest;

typedef struct blockarrayinfo {
	BlockDistArray * block_array;
	Name name;
	void * buffer;
	int lower;
	int higher;
	int request;
	int request_id;
	int no_reads;
} BlockArrayInfo;

typedef struct cyclicarrayinfo {
	CyclicDistArray * cyclic_array;
	Name name;
	void * buffer;
	int lower;
	int higher;
	int request;
	int request_id;
	int no_reads;
} CyclicArrayInfo;

typedef struct arrayholder {
	BlockDistArray ** block_arrays;
	int no_block_arrays;
	CyclicDistArray ** cyclic_arrays;
	int no_cyclic_arrays;
} ArrayHolder;

int compare_array_b(const void * _info1, const void * _info2);
BlockDistArray * locate_array_b(ArrayHolder * _holder, Name _name);
CyclicDistArray * locate_array_c(ArrayHolder * _holder, Name _name);
int compare_block_pointers(const void * _pointer1, const void * _pointer2);
int compare_cyclic_pointers(const void * _pointer1, const void * _pointer2);
void print_b_arrays_holder(ArrayHolder * _holder);

#endif
