
/******************************************************************************
 * Tlib
 *
 * Copyright (C) 2001, Thomas Rauber
 * All rights reserved 
 *
 * changed by: Sascha Hunold <hunold@uni-bayreuth.de>
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpi.h>

#include "tlib.h"

/**
 * reverses char string in place 
 */
void reverse (char s[]) {
  int c, i, j;
  for (i=0, j=strlen(s)-1; i<j; i++,j--) {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

/**
 * integer to ASCII 
 */
void p_itoa (int n, char s[]) {
  int i,sign;
  if ((sign = n) < 0)
     n = -n;
  i = 0;
  do {
    s[i++] = n % 10 + '0';
  } while ((n /= 10) > 0);
  if (sign < 0)
    s[i++] = '-';
  s[i] = '\0';
  reverse(s);
}


void print_T_Descr (T_Descr *pdescr) {
  int i, me = pdescr->my_rank;
  T_Descr *pparent;
 
  printf("------------------------------------------------\n");
  printf("p%d: printing T descriptor\n",me);
  printf("p%d: group name: %s \t group size: %d \t my_pos_in_group: %d\n",
                       me, pdescr->name,pdescr->size,pdescr->my_rank_local);
  printf("p%d: first rank: %d \t my rank: %d \t my group nr: %d\n",
                       me, pdescr->first_rank,pdescr->my_rank,
                       pdescr->my_group_nr);
  printf("p%d: members of my group: ",me);
  for (i=0; i<pdescr->size; i++)
     printf("%d ",pdescr->my_group[i]);
  printf("\n");
  printf("p%d: number of groups: %d \t group sizes: ",
             me, pdescr->no_of_groups);
  for (i=0; i<pdescr->no_of_groups; i++)
     printf("%d ",pdescr->group_sizes[i]);
  printf("\n"); 
  if (pdescr->parent == NULL) printf("no parent group\n");
  else {
    pparent = pdescr->parent;
    printf("p%d: parent group: %s with size %d:", 
           me, pparent->name, pparent->size);
    for (i=0; i < pparent->size; i++)
      printf(" %d ", pparent->my_group[i]);
    printf("\n");
  }
  printf("------------------------------------------------\n");
}


void copy_T_Descr (T_Descr *source, T_Descr *target) {
  int i,j,k,g,*start;
  char *x;
  
  x=malloc(sizeof(char)*strlen(source->name));

  x = strcpy(target->name,source->name);
  target->size = source->size;
  target->my_rank = source->my_rank;
  target->my_rank_local = source->my_rank_local;
  target->first_rank = source->first_rank;
  target->my_group_nr = source->my_group_nr;
  target->no_of_groups = g = source->no_of_groups;
  target->group_sizes = (int *) malloc (g * sizeof(int));
  target->group_leaders = (int *) malloc (g * sizeof(int));
  target->all_groups = (int **) malloc (g * sizeof(int *));
  for (i=0; i<source->no_of_groups; i++) {
    target->group_sizes[i] = source->group_sizes[i];
    target->group_leaders[i] = source->group_leaders[i];
    target->all_groups[i] = (int *) 
                       malloc (source->group_sizes[i] * sizeof(int));
    for (k=0; k<source->group_sizes[i]; k++)
       target->all_groups[i][k] = source->all_groups[i][k];
  }
  target->my_group = target->all_groups[target->my_group_nr];

    
  MPI_Comm_dup(source->comm, &(target->comm));
  MPI_Comm_group(target->comm, &(target->mpigroup));
  target->parent = source->parent;
}


void T_Init (int *argc, char ***argv, MPI_Comm comm, T_Descr *pdescr)
{ 
  int myrank, i, p;

  MPI_Init(argc, argv);
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &p); 

  (*pdescr).status = VALID;
  (*pdescr).name = (char *) malloc (sizeof(char) * 2);
  (*pdescr).name[0] ='0';                /* name of first group is "0" */
  (*pdescr).name[1] ='\0';               
  (*pdescr).size = p;
  (*pdescr).my_rank = myrank;
  (*pdescr).my_rank_local = myrank;
  (*pdescr).first_rank = 0;
  (*pdescr).no_of_groups = 1;
  (*pdescr).my_group_nr = 0;
  (*pdescr).all_groups = (int **) malloc (sizeof(int *) * 1);
  (*pdescr).all_groups[0] = (int *) malloc (sizeof(int) * p);
  (*pdescr).my_group = (*pdescr).all_groups[0];
  for (i=0; i<p; i++) ((*pdescr).my_group)[i] = i ;
  (*pdescr).group_sizes = (int *) malloc (sizeof(int) * 1);
  (*pdescr).group_sizes[0] = p;
  (*pdescr).group_leaders = (int *) malloc (sizeof(int) * 1);
  (*pdescr).group_leaders[0] = 0;
  MPI_Comm_dup(comm,&((*pdescr).comm));
  MPI_Comm_group (comm, &((*pdescr).mpigroup));
  (*pdescr).parent = NULL;           /* there is no parents group */
#ifdef TWOL_DEBUG
  print_T_Descr(pdescr);
#endif	
}


void T_Finalize() {
  MPI_Finalize();
}


int T_Split_GrpParfor(int no_of_groups, 
		      T_Descr *descr, 
		      T_Descr *pdescr,
		      float *per)
{ 
  int newrank, myrank=descr->my_rank, p, mygroup,i,j,k,total;
  int *g, *start;
  float per_sum = 0.0;
  MPI_Comm newcomm;
  char *dummy,*i2a;

  for (i=0; i<no_of_groups; i++) per_sum += per[i];
  if (per_sum > 1.0) {
     printf("illegal split operation: wrong split specification\n");
     copy_T_Descr(descr, pdescr);
     pdescr -> status = ERROR;
     return ERROR_ILLEGAL_SPLIT;
  }
  p = descr->size;
  if (no_of_groups > p) {
     printf("not enough processors available for splitting of group %s\n",
             descr->name);
     copy_T_Descr(descr, pdescr);
     pdescr -> status = INVALID;  
     return ERROR_INVALID_ARGS;
  }
  pdescr -> status = VALID;

  /* select group sizes according to specification */
  g = split_group(p, per, no_of_groups);

  /* determine the proc.id of the first processor in each group: */
  /* additional entry: start[no_of_groups] = p for similar cond  */
  start = (int *) malloc ((no_of_groups+1) * sizeof(int));
  total = descr->first_rank;
  for (i=0; i<no_of_groups; i++) { 
    start[i] = total; total = total + g[i];
  }
  start[i] = total;
  /* printf("start: %d %d %d \n",start[0],start[1], start[2]); */
  pdescr->group_leaders = start;

  /* additional entries in T decriptor */
  pdescr->group_sizes = g;
  pdescr->no_of_groups = no_of_groups;

  /* each proc determines its own group number */
  MPI_Comm_dup(descr->comm,&newcomm);
  for (i=0; i<no_of_groups; i++) 
    if ((myrank >= start[i]) && (myrank < start[i+1])) {
      mygroup = pdescr->my_group_nr = i; 
      pdescr->size = g[i];
      break; 
  }
  pdescr->first_rank = pdescr->group_leaders[pdescr->my_group_nr];
  MPI_Comm_split (descr->comm, mygroup, myrank, &newcomm);
  MPI_Comm_rank (descr->comm, &newrank);

  /* determine the new group name */
  dummy = (char *) malloc (sizeof (char) * 
                           (strlen(descr->name) + 2 + no_of_groups/10));
  i2a = (char *) malloc (sizeof (char) * mygroup/10 + 1);
  strcpy (dummy, descr->name);
  p_itoa(mygroup,i2a);
  dummy = strcat(dummy,".");
  (*pdescr).name =  strcat(dummy,i2a);

  /* each processor sets its own id */
  (*pdescr).my_rank = myrank;
  (*pdescr).my_rank_local = myrank - pdescr->group_leaders[pdescr->my_group_nr];

  /* determine the all sibling group (= groups on the same level)     */
  /*           and the corresponding proc ids                         */
  (*pdescr).all_groups = (int **) malloc (no_of_groups * sizeof(int *));
  for (i=0; i<no_of_groups; i++) {
    (*pdescr).all_groups[i] = (int *) malloc (g[i] * sizeof(int));
    for (k=0,j=start[i]; j<start[i+1]; j++,k++) 
       ((*pdescr).all_groups)[i][k] = j;
    if (mygroup == i) {
      (*pdescr).size = g[i];
      (*pdescr).my_group = (*pdescr).all_groups[i];
    }
  }
  MPI_Comm_dup(newcomm,&((*pdescr).comm));
  MPI_Comm_group ((*pdescr).comm, &((*pdescr).mpigroup));
  
  /* set the pointer to the parents twoL descriptor */
  ((*pdescr).parent) = descr;
  
  /* print info for new group partition */
  /* print_T_Descr(pdescr); */
  return 0;
}


int T_Split_Grp(T_Descr *descr,
		T_Descr *pdescr,
		float per1, float per2)
{ 
  float per[2];
  int ret;
  
  per[0] = per1;
  per[1] = per2;

  if ( per[0] + per[1] > 1.0) {
    printf("illegal split operation: wrong split specification\n");
    copy_T_Descr (descr, pdescr);
    pdescr -> status = ERROR;
    ret = ERROR_ILLEGAL_SPLIT;
  }
  else {
    ret = T_Split_GrpParfor(2,descr,pdescr,per); 
  }
  return ret;
}


int T_Split_GrpExpl( T_Descr *descr,
                  T_Descr *pdescr,
                  int _color, int _key)
{
  int 
    p = descr->size, 
    /* we must take local rank -> now we can rearrange the groups */
    myrank = descr->my_rank_local, 
    *color_field, 
    *different_colors, 
    *key_field, 
    *group_size, 
    *start, 
    *ranks, 
    smallest_key_pos, 
    my_group_start, 
    *group,
    my_group_nr, 
    current_col, 
    n_diff_colors = 0, 
    total_procs = 0;

  MPI_Comm newcomm;
  char *dummy,*i2a;
  int i,j,k,exists;

  MPI_Comm_dup(descr->comm,&newcomm);
  MPI_Comm_split (descr->comm, _color, _key, &newcomm);
  MPI_Comm_dup(newcomm, &pdescr->comm);
  MPI_Comm_group( pdescr->comm, &pdescr->mpigroup);
  pdescr->parent = descr;

  color_field      = (int *) malloc (sizeof(int) * p);
  different_colors = (int *) malloc (sizeof(int) * p);
  key_field        = (int *) malloc (sizeof(int) * p);
  group_size       = (int *) malloc (sizeof(int) * p);
  start            = (int *) malloc (sizeof(int) * p);
  ranks            = (int *) malloc (sizeof(int) * p);
  group            = (int *) malloc (sizeof(int) * p);

  for (i=0; i<p; i++) 
    group_size[i] = 0;
  
  /* make color and key entries available  for each processor */
  /* Mattias changed this 050505, Scampi and tlib bugs with allgather */
  
  //MPI_Allgather( &_color, 1, MPI_INT, color_field, 1, MPI_INT, descr->comm);
  MPI_Gather(&_color, 1, MPI_INT, color_field, 1, MPI_INT, 0, descr->comm);
  MPI_Bcast(color_field, p, MPI_INT, 0, descr->comm);
  //MPI_Allgather( &_key, 1, MPI_INT, key_field, 1, MPI_INT, descr->comm);
  MPI_Gather(&_key, 1, MPI_INT, key_field, 1, MPI_INT, 0, descr->comm);
  MPI_Bcast(key_field, p, MPI_INT, 0, descr->comm);
  /* ranks are required to order processors if keys are equal */
  //MPI_Allgather( &myrank, 1, MPI_INT, ranks, 1, MPI_INT, descr->comm);
  MPI_Gather(&myrank, 1, MPI_INT, ranks, 1, MPI_INT, 0, descr->comm);
  MPI_Bcast(ranks, p, MPI_INT, 0, descr->comm);

  /* collect in diff_colors the different colors specified     */
  /* n_diff_color counts the number of differnt colors         */
  /*                         = number of sibling groups        */
  /* array group_sizes counts the size of the different groups */
  for (i=0; i<p; i++) {
    exists = 0;
    current_col = color_field[i];
    for (j=0; j<n_diff_colors; j++)
      if (current_col == different_colors[j]) {
	exists = 1;
	group_size[j]++;
	group[i] = j;
	break;
      }
    
    if (! exists) { /* new color found */
      different_colors[n_diff_colors] = current_col;
      group_size[n_diff_colors] = 1;
      group[i] = n_diff_colors;
      n_diff_colors++;
    }

    if( _color == current_col ) 
      my_group_nr = group[i];
  }

  /* find smallest key with group color */
  /* -> determines the starting processor of this group */
  for(i=0; i<n_diff_colors; i++) {
    smallest_key_pos = -1;
    for(k=0; k<p; k++) {
      if( group[k] != i )
	continue;
      if( (smallest_key_pos == -1) || (key_field[k] < key_field[smallest_key_pos]) || 
	  (key_field[k] == key_field[smallest_key_pos] && k < smallest_key_pos) ) {
	smallest_key_pos = k;
      }
    }
    start[i] = ranks[smallest_key_pos];
  }


  for (i=0; i<n_diff_colors; i++) {
    total_procs += group_size[i];
  }
  
  if (total_procs > p) {
    fprintf(stderr, "p%d: too few processors available for splitting group %s\n", 
	    descr->my_rank, descr->name);
  }

  pdescr->status = VALID;
  pdescr->no_of_groups = n_diff_colors;
  pdescr->my_rank = descr->my_rank;
  pdescr->group_sizes = group_size;
  pdescr->my_group_nr = my_group_nr;
  pdescr->size = group_size[my_group_nr];
  pdescr->group_leaders = start;
  MPI_Comm_rank( newcomm, &pdescr->my_rank_local );

  /* construct the array of groups */
  pdescr->all_groups = (int **) malloc (n_diff_colors * sizeof(int *));
  for (i=0; i<n_diff_colors; i++) {
    (*pdescr).all_groups[i] = (int *) malloc (group_size[i] * sizeof(int));
    for (j=0, k=0; j<p; j++)
      if( group[j] == i )
	((*pdescr).all_groups)[i][k++] = j;
    if (my_group_nr == i) {
      (*pdescr).my_group = (*pdescr).all_groups[i];
    }
  }

  pdescr->first_rank = start[my_group_nr];

  dummy = (char *) malloc((strlen(descr->name) + 2 + n_diff_colors/10));
  i2a = (char *) malloc (my_group_nr/10 + 1);
  strcpy (dummy, descr->name);
  p_itoa(my_group_nr,i2a);
  dummy = strcat(dummy,".");
  pdescr->name =  strcat(dummy,i2a);

  /* cleanup */
  free(color_field);
  free(different_colors);
  free(key_field);
  free(ranks);
  free(group);

  return 0;
}



int T_Par(void * (*f1)(void *, MPI_Comm, T_Descr *), 
	  void * parg1, 
	  void ** pres1,
	  void * (*f2)(void *, MPI_Comm, T_Descr *),
	  void * parg2, 
	  void ** pres2,
	  T_Descr *descr)
{
  int ret = ERROR_NONE;

  /*  print_T_Descr(descr);                */
  *pres1 = *pres2 = NULL;
  if (descr->status == ERROR) { 
    fprintf(stderr, "invalid split operation\n"); 
    ret = ERROR_INVALID_ARGS;
  }
  else if (descr->status == INVALID) {
    fprintf(stderr, "executing parallel composition consecutively\n"); 
    *pres1 = f1 (parg1, (descr->comm), descr);
    *pres2 = f2 (parg2, (descr->comm), descr);
    ret = ERROR_INVALID_ARGS;
  }
  else { 
    if (descr->my_group_nr == 0) {
      *pres1 = f1 (parg1, (descr->comm), descr);
    }
    else if (descr->my_group_nr == 1) {
      *pres2 = f2 (parg2, (descr->comm), descr);
    }
    else {
      fprintf(stderr, "proc %d not participating in computation\n",descr->my_rank);
    }
    
    if (descr->parent != NULL) {
      MPI_Barrier ((*descr->parent).comm);
    }
    else {
      fprintf(stderr, "p%d: internal error: Twol descriptor NULL\n",descr->my_rank);
      ret = ERROR_INTERNAL_NULL;
    }
  }
  return ret;
}


int T_Parfor(void * (*f[])(void *, MPI_Comm, T_Descr *),
             void * parg[],
             void * pres[],
             T_Descr *descr)
{ 
  int i, ret = ERROR_NONE;

  if (descr->status == ERROR) { 
    fprintf(stderr, "invalid split operation\n"); 
    ret = ERROR_INVALID_ARGS;
  }
  else {
    for (i=0; i<descr->no_of_groups; i++) {
      pres[i] = NULL;
    }
    
    if (descr->status == INVALID) {
      fprintf(stderr, "executing parallel composition consecutively\n"); 
      for (i=0; i<descr->no_of_groups; i++)
	pres[i] = (f[i])(parg[i], descr->comm, descr);
      ret = ERROR_INVALID_ARGS;
    }
    else { 
      i = descr->my_group_nr;
#ifdef TWOL_DEBUG
      fprintf(stderr, "p%d: in TwolParfor calling %d\n",descr->my_rank, i);
      print_T_Descr(descr);
#endif		
      pres[i] = (f[i])(parg[i], descr->comm, descr);
      if (descr->parent != NULL) {
	MPI_Barrier ((*descr->parent).comm);
      }
      else {
	fprintf(stderr, "p%d: internal error: Twol descriptor NULL\n",descr->my_rank);
	fflush(stderr);
      }
    }
  }
  return ret;
}

int T_GetMyGroupSize (T_Descr *descr)
{
  return (descr->group_sizes[descr->my_group_nr]);
}


int T_GetMyRank (T_Descr *descr)
{
  return descr->my_rank;
}


int T_GetMyRankLocal (T_Descr *descr)
{
  return descr->my_rank_local;
}

int *T_GetGroupLeaders (T_Descr *descr)
{
  int *rval, i, size = descr->no_of_groups;

  rval = (int *) malloc (sizeof(int) * size);
  for (i=0; i<size; i++)
    rval[i] = descr->group_leaders[i];
  return rval;
}

MPI_Comm T_GetComm (T_Descr *descr)
{
  return descr->comm;
}

T_Descr * T_GetParent (T_Descr *descr)
{
  return descr->parent;
}

int T_GetMyGroupNo (T_Descr *descr)
{
  return descr->my_group_nr;
}

int * T_GetMyGroup (T_Descr *descr)
{
  return descr->my_group;
}


int T_ReDistrRepl (void * source1, void * target1, int size1,
		   void * source2, void * target2, int size2,
		   T_Descr *parent, T_Descr *child)
{
   int *first;
   
   first = T_GetGroupLeaders(child);
   if (T_GetMyRank(child) == first[0]) 
      memcpy(target1, source1, size1);
   if (T_GetMyRank(child) == first[1]) 
      memcpy(target2, source2, size2);
   MPI_Bcast(target1, size1, MPI_BYTE, first[0], parent->comm);
   MPI_Bcast(target2, size2, MPI_BYTE, first[1], parent->comm);
   return 0;
}

int *split_group(int p, float *per, int subg_num) 
{
  int *g;
  int count, i, total, pos;
  
  g = (int*)malloc(subg_num*sizeof(int));

  for (i=0, total = 0; i<subg_num; i++) {
    g[i] = per[i]*p; total = total + g[i];
  }

  count = 0; pos = 0;
  while( total++ < p ) {
    if( count % 2 == 0 ) 
      i = (subg_num - 1 - pos) % subg_num;
    else 
      i = pos++;
    g[i]++;
    count++;
  }

  return g;
}









