#include "tree_builders.h"
#include "NestStep.h"
#include <stdio.h>
#include <math.h>

int tree_style = 2;
double bval = 0.5;
int dval = 4;

int change_tree_dary(int _dval)
{
	tree_style = DARY;
	dval = _dval;

	return 0;
}

int change_tree_flat(void)
{
	tree_style = FLAT;

	return 0;
}

int change_tree_bin(double _bval)
{
	tree_style = BIO;
	bval = _bval;

	return 0;
}

int rebuild_tree(void)
{
	free_Tree();
	new_Tree();

	return 0;
}

int build_tree(int _nodes, int _rank, int * _parent, 
			   int * _no_children, int ** _children)
{
	switch(tree_style) {
		case DARY:
			build_d_tree(0, _nodes - 1, _rank, _parent, _no_children, _children, dval);
			break;
		case FLAT:
			build_flat_tree(_nodes, _rank, _parent, _no_children, _children);
			break;
		case BIO:
			skewed_tree_init(_nodes, bval, _rank, _parent, _no_children, _children);
			break;
		default:
			build_d_tree(0, _nodes - 1, _rank, _parent, _no_children, _children, dval);
	}
	return 0;
}

void skewed_tree_init(int nodes, double ratio, int rank, int * parent, 
					  int * no_children, int ** children)
{
	*no_children = 0;
	if(rank >= 1)
		*parent = 0;
	else
		*parent = -1;
	if(nodes > 1)
		skewed_tree(1, nodes - 1, ratio, rank, parent, no_children, children);
}

void skewed_tree(int _lower, int _higher, double _ratio, int _rank, 
				 int * _parent, int * _no_children, int ** _children)
{
	int amount;
	int amount_lower;
	int lower_low;
	int lower_high;
	int upper_low;
	int upper_high;

	amount = _higher - _lower + 1 - 1;
	amount_lower = (int)round(_ratio * (double)amount);
	lower_low = _lower + 1;
	lower_high = _lower + amount_lower;
	upper_low = lower_high + 1;
	upper_high = _higher;

//#ifdef TREE_DEBUG
//	printf("%d: amount: %d: amount_lower: %d lower: %d:%d upper: %d:%d\n",
//		_rank, amount, amount_lower,
//		lower_low, lower_high,
//		upper_low, upper_high);
//#endif
	
	if(_rank < _lower)
	{
		(*_children)[*_no_children] = _lower;
		*_no_children += 1;
		if(amount - amount_lower > 0)
			skewed_tree(upper_low, upper_high, _ratio, _rank, _parent, _no_children, _children);
	}
	else if(_rank == _lower)
	{
		if(amount_lower > 0)
			skewed_tree(lower_low, lower_high, _ratio, _rank, _parent, _no_children, _children);
	}
	else if(_rank >= lower_low && _rank <= lower_high)
	{
		*_parent = _lower;
		skewed_tree(lower_low, lower_high, _ratio, _rank, _parent, _no_children, _children);
	}
	else if(_rank >= upper_low && _rank <= upper_high)
	{
		skewed_tree(upper_low, upper_high, _ratio, _rank, _parent, _no_children, _children);
	}
	else
	{
		printf("Error in skewed_tree.\n");
	}
}

void build_flat_tree(int _nodes, int _rank, int * _parent, 
					 int * _no_children, int ** _children)
{
	int i;
	if(_rank == 0)
	{
		*_parent = -1;
		*_no_children = _nodes - 1;
		for(i = 0; i < _nodes - 1; i++)
			(*_children)[i] = i + 1;
	}
	else
	{
		*_parent = 0;
		*_no_children = 0;
	}
}

void build_d_tree(int lower, int higher, int rank, int * parent,
				  int * no_children, int ** children, int d)
{
	int rem;
	int ratio;
	int no_nodes;
	int i;
	int start, end;

	no_nodes = higher - lower + 1;
	rem = (no_nodes - 1) % d;
	ratio = (no_nodes - 1) / d;

	if(rank == 0)
		*parent = -1;
	if(no_nodes <= 1)
	{
		*no_children = 0;
		return;
	}

	if(ratio >= 1)
	{
		if(rank == lower)
		{
			*no_children = d;
		}
		else
		{
			*parent = lower;
			*no_children = 0;
		}

		end = lower;
		for(i = 0; i < d; ++i)
		{
			start = end + 1;
			end = start + ratio - 1 + ((i < rem) ? 1 : 0);

			if(rank == lower)
				(*children)[i] = start;
			else if(rank >= start && rank <= end)
				build_d_tree(start, end, rank, parent, no_children, children, d);
		}
	}
	else
	{
		if(rank == lower)
		{
			*no_children = no_nodes - 1;
			for(i = 0; i < no_nodes - 1; ++i)
				(*children)[i] = lower + 1 + i;
		}
		else
		{
			*parent = lower;
			*no_children = 0;
		}
	}
}

//int round(double _d)
//{
//	double delta_c;
//	double delta_f;
//
//	delta_c = ceil(_d) - _d;
//	delta_f = _d - floor(_d);
//
//	if(delta_c < delta_f)
//		return (int)ceil(_d);
//	else
//		return (int)floor(_d);
//}
